import wgpu  # only for flags/enums

from ...utils import array_from_shadertype


class RenderTexture:
    """Class used internally to store a render texture and meta data."""

    def __init__(self, format):
        self.format = format
        self.texture = None
        self.texture_view = None
        self.size = (0, 0, 0)

    def set_texture_view(self, texture_view):
        """Set from a texture view. Intended when the texture comes
        from elsewhere. Make sure it matches the format!
        """
        self.texture = None
        self.texture_view = texture_view
        self.size = self.texture_view.size

    def ensure_size(self, device, size):
        """Make sure that the texture has the given size. If necessary,
        recreates the texture and texture-view objects.
        """
        if size != self.size:
            self.size = size
            usage = wgpu.TextureUsage.RENDER_ATTACHMENT | wgpu.TextureUsage.COPY_SRC
            if self.format.startswith(("rgb", "bgr")):
                usage |= wgpu.TextureUsage.SAMPLED
            self.texture = device.create_texture(
                size=size, usage=usage, dimension="2d", format=self.format
            )
            self.texture_view = self.texture.create_view()

    @property
    def bytes_per_pixel(self):
        """The number of bytes per pixel."""
        format_map = {
            "depth24plus_stencil8": 4,
            "depth24plus": 3,  # ?
            "depth32": 4,
            "r8": 1,
            "r16": 2,
            "r32": 4,
            "rg8": 2,
            "rg16": 4,
            "rg32": 8,
            "rgba8": 4,
            "rgba16": 8,
            "rgba32": 16,
        }
        for key, val in format_map.items():
            if self.format.startswith(key):
                return val
        else:
            raise ValueError(f"Could not determine bbp of {self.format}")


class PostProcessingStep:
    """
    Utility object to render one texture to another, as a post-processing step.
    """

    def __init__(self, device, *, vertex_shader=None, fragment_shader, uniform_type):
        self._device = device
        self._vertex_shader = (
            default_vertex_shader if vertex_shader is None else vertex_shader
        )
        self._fragment_shader = fragment_shader
        self._hash = None

        self._uniform_type = uniform_type
        self._uniform_data = array_from_shadertype(self._uniform_type)
        self._uniform_buffer = self._device.create_buffer(
            size=self._uniform_data.nbytes,
            usage=wgpu.BufferUsage.UNIFORM | wgpu.BufferUsage.COPY_DST,
        )

        self._sampler = self._device.create_sampler(
            label="render sampler",
            mag_filter="nearest",
            min_filter="nearest",
        )

    def render(self, src, dst):
        """Render one texture to another. If the src.texture_view and the dst.format
        have not changed, the pipeline is re-used.
        """
        assert isinstance(src, RenderTexture)
        assert isinstance(dst, RenderTexture)
        device = self._device

        # Recreate pipeline?
        hash = src.size, id(src.texture_view), dst.format
        if hash != self._hash:
            self._hash = hash
            self._create_pipeline(
                self._vertex_shader, self._fragment_shader, src.texture_view, dst.format
            )

        command_encoder = device.create_command_encoder()

        # Copy data to tmp buffer
        tmp_buffer = device.create_buffer_with_data(
            data=self._uniform_data,
            usage=wgpu.BufferUsage.COPY_SRC,
        )
        command_encoder.copy_buffer_to_buffer(
            tmp_buffer, 0, self._uniform_buffer, 0, self._uniform_data.nbytes
        )

        render_pass = command_encoder.begin_render_pass(
            color_attachments=[
                {
                    "view": dst.texture_view,
                    "resolve_target": None,
                    "load_value": (0, 0, 0, 0),  # LoadOp.load or color
                    "store_op": wgpu.StoreOp.store,
                }
            ],
            depth_stencil_attachment=None,
            occlusion_query_set=None,
        )
        render_pass.set_pipeline(self._render_pipeline)
        render_pass.set_bind_group(0, self._bind_group, [], 0, 99)
        render_pass.draw(4, 1)
        render_pass.end_pass()
        device.queue.submit([command_encoder.finish()])

    def _create_pipeline(
        self, vertex_shader, fragment_shader, src_texture_view, dst_format
    ):
        device = self._device

        vs_module = device.create_shader_module(code=vertex_shader)
        fs_module = device.create_shader_module(code=fragment_shader)

        binding_layouts = [
            {
                "binding": 0,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "sampler": {"type": wgpu.SamplerBindingType.filtering},
            },
            {
                "binding": 1,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "texture": {
                    "sample_type": wgpu.TextureSampleType.float,
                    "view_dimension": wgpu.TextureViewDimension.d2,
                    "multisampled": False,
                },
            },
            {
                "binding": 2,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "buffer": {"type": wgpu.BufferBindingType.uniform},
            },
        ]
        bindings = [
            {"binding": 0, "resource": self._sampler},
            {"binding": 1, "resource": src_texture_view},
            {
                "binding": 2,
                "resource": {
                    "buffer": self._uniform_buffer,
                    "offset": 0,
                    "size": self._uniform_data.nbytes,
                },
            },
        ]

        bind_group_layout = device.create_bind_group_layout(entries=binding_layouts)
        pipeline_layout = device.create_pipeline_layout(
            bind_group_layouts=[bind_group_layout]
        )
        self._bind_group = device.create_bind_group(
            layout=bind_group_layout, entries=bindings
        )

        self._render_pipeline = device.create_render_pipeline(
            layout=pipeline_layout,
            vertex={
                "module": vs_module,
                "entry_point": "main",
                "buffers": [],
            },
            primitive={
                "topology": wgpu.PrimitiveTopology.triangle_strip,
                "strip_index_format": wgpu.IndexFormat.uint32,
            },
            depth_stencil=None,
            multisample=None,
            fragment={
                "module": fs_module,
                "entry_point": "main",
                "targets": [
                    {
                        "format": dst_format,
                        "blend": {
                            "alpha": (
                                wgpu.BlendFactor.one,
                                wgpu.BlendFactor.zero,
                                wgpu.BlendOperation.add,
                            ),
                            "color": (
                                wgpu.BlendFactor.src_alpha,
                                wgpu.BlendFactor.one_minus_src_alpha,
                                wgpu.BlendOperation.add,
                            ),
                        },
                    }
                ],
            },
        )


class SSAAPostProcessingStep(PostProcessingStep):
    """A texture renderer that supports SSAA."""

    def __init__(self, device):
        super().__init__(
            device, fragment_shader=ssaa_fragment_shader, uniform_type=ssaa_uniform_type
        )

    def render(self, src, dst):

        # Get factor between texture sizes
        factor_x = src.size[0] / dst.size[0]
        factor_y = src.size[1] / dst.size[1]
        factor = (factor_x + factor_y) / 2

        if factor > 1:
            # src has higher res, we can do ssaa
            sigma = 0.5 * factor
            support = min(5, int(sigma * 3))
        else:
            # src has lower res. smooth those pixels a bit
            sigma = 1
            support = 2

        self._uniform_data["size"] = src.size[:2]
        self._uniform_data["sigma"] = sigma
        self._uniform_data["support"] = support

        super().render(src, dst)


# %% Shaders

ssaa_uniform_type = dict(
    size=("float32", 2),
    sigma=("float32",),
    support=("int32",),
)


default_vertex_shader = """

struct VertexOutput {
    [[location(0)]] texcoord: vec2<f32>;
    [[builtin(position)]] pos: vec4<f32>;
};

[[stage(vertex)]]
fn main([[builtin(vertex_index)]] index: u32) -> VertexOutput {
    let positions = array<vec2<f32>, 4>(vec2<f32>(0.0, 1.0), vec2<f32>(0.0, 0.0), vec2<f32>(1.0, 1.0), vec2<f32>(1.0, 0.0));
    let pos = positions[index];
    var out: VertexOutput;
    out.texcoord = vec2<f32>(pos.x, 1.0 - pos.y);
    out.pos = vec4<f32>(pos * 2.0 - 1.0, 0.0, 1.0);
    return out;
}

"""

ssaa_fragment_shader = """
struct VertexOutput {
    [[location(0)]] texcoord: vec2<f32>;
    [[builtin(position)]] pos: vec4<f32>;
};

[[block]]
struct Render {
    size: vec2<f32>;
    sigma: f32;
    support: i32;
};

[[group(0), binding(0)]]
var r_sampler: sampler;
[[group(0), binding(1)]]
var r_tex: texture_2d<f32>;
[[group(0), binding(2)]]
var u_render: Render;

[[stage(fragment)]]
fn main(in: VertexOutput) -> [[location(0)]] vec4<f32> {
    // Get info about the smoothing
    let sigma = u_render.sigma;
    let support = min(5, u_render.support);

    // Determine distance between pixels in src texture
    let stepp = vec2<f32>(1.0 / u_render.size.x, 1.0 / u_render.size.y);
    // Get texcoord, and round it to the center of the source pixels.
    // Thus, whether the sampler is linear or nearest, we get equal results.
    let ori_coord = in.texcoord.xy;
    let ref_coord = vec2<f32>(vec2<i32>(ori_coord / stepp)) * stepp + 0.5 * stepp;

    // Convolve. Here we apply a Gaussian kernel, the weight is calculated
    // for each pixel individually based on the distance to the actual texture
    // coordinate. This means that the textures don't even need to align.
    var val: vec4<f32> = vec4<f32>(0.0, 0.0, 0.0, 0.0);
    var weight: f32 = 0.0;
    for (var y:i32 = -support; y <= support; y = y + 1) {
        for (var x:i32 = -support; x <= support; x = x + 1) {
            let coord = ref_coord + vec2<f32>(f32(x), f32(y)) * stepp;
            let dist = length((ori_coord - coord) / stepp);  // in src pixels
            let t = dist / sigma;
            let w = exp(-0.5 * t * t);
            val = val + textureSample(r_tex, r_sampler, coord) * w;
            weight = weight + w;
        }
    }
    return val / weight;
}
"""



from ..objects._base import ResourceContainer


class Material(ResourceContainer):

    uniform_type = {}

    def _wgpu_get_pick_info(self, pick_value):
        """Given a 4 element tuple, sampled from the pick texture,
        return info about what was picked in the object. The first value
        refers to the object id. The use of the remaining values differs
        per material.
        """
        # Note that this is a private friend-method of the renderer.
        return {}



import numpy as np

from ._buffer import Resource, STRUCT_FORMAT_ALIASES

# todo: what to do about these enums from wgpu. Copy them over?


class Texture(Resource):
    """A base texture wrapper that can be implemented for numpy, ctypes arrays,
    or any other kind of array.

    Parameters:
        data (array, optional): Array data of any type that supports the
            buffer-protocol, (e.g. a bytes or numpy array). If not given
            or None, nbytes and nitems must be provided. The data is
            copied if it's float64 or not contiguous.
        dim (int): The dimensionality of the array (1, 2 or 3).
        usage: The way(s) that the texture will be used. Default "SAMPLED",
            set/add "STORAGE" if you're using it as a storage texture
            (see wgpu.TextureUsage).
        size (3-tuple): The extent ``(width, height, depth)`` of the array.
            If not given or None, it is derived from dim and the shape of
            the data. By creating a 2D array with ``depth > 1``, a view can
            be created with format 'd2_array' or 'cube'.
        format (enum str): the GPU format of texture. Must be a value from
            wgpu.TextureFormat. By default it is derived from the data. Set when
            data is not given or when you want to overload the derived value.
    """

    def __init__(self, data=None, *, dim, usage="SAMPLED", size=None, format=None):
        self._rev = 0
        # The dim specifies the texture dimension
        assert dim in (1, 2, 3)
        self._dim = int(dim)
        # The size specifies the size on the GPU (width, height, depth)
        self._size = ()
        self._format = None if format is None else str(format)
        self._nbytes = 0
        # The actual data (optional)
        self._data = None
        self._pending_uploads = []  # list of (offset, size) tuples

        size = None if size is None else (int(size[0]), int(size[1]), int(size[2]))

        if data is not None:
            mem = memoryview(data)
            if mem.format == "d":
                raise ValueError("Float64 data is not supported, use float32 instead.")
            self._data = data
            self._mem = mem
            self._nbytes = mem.nbytes
            self._size = self._size_from_data(mem, dim, size)
            self.update_range((0, 0, 0), self._size)
        elif size is not None and format is not None:
            self._size = size
        else:
            raise ValueError(
                "Texture must be instantiated with either data or size and format."
            )

        # Determine usage
        if isinstance(usage, str):
            usages = usage.upper().replace(",", " ").replace("|", " ").split()
            assert usages
            self._usage = "|".join(usages)
        else:
            raise TypeError("Texture usage must be str.")

    @property
    def rev(self):
        """An integer that is increased when update_range() is called."""
        return self._rev

    def get_view(self, **kwargs):
        """Get a new view on the this texture."""
        return TextureView(self, **kwargs)

    @property
    def dim(self):
        """The dimensionality of the texture (1, 2, or 3)."""
        return self._dim

    @property
    def usage(self):
        """The texture usage flags as a string (compatible with the
        wgpu.TextureUsage enum).
        """
        return self._usage

    @property
    def data(self):
        """The data for this texture. Can be None if the data only
        exists on the GPU.

        Note: the data is the same reference that was given to
        instantiate this object, but this may change.
        """
        return self._data

    @property
    def mem(self):
        """The data for this buffer as a memoryview. Can be None if
        the data only exists on the GPU.
        """
        return self._mem

    @property
    def nbytes(self):
        """Get the number of bytes in the texture."""
        return self._nbytes

    @property
    def size(self):
        """The size of the texture as (width, height, depth).
        (always a 3-tuple, regardless of the dimension).
        """
        return self._size

    @property
    def format(self):
        """The texture format as a string (compatible with the
        wgpu.TextureFormat enum).
        """
        if self._format is not None:
            return self._format
        elif self.usage == "UNIFORM":
            return None
        elif self.data is not None:
            self._format = format_from_memoryview(self.mem, self.size)
            return self._format
        else:
            raise ValueError("Texture has no data nor format.")

    def update_range(self, offset, size):
        """Mark a certain range of the data for upload to the GPU.
        The offset and (sub) size should be (width, height, depth)
        tuples. Numpy users beware that an arrays shape is (height, width)!
        """
        # Check input
        assert isinstance(offset, tuple) and len(offset) == 3
        assert isinstance(size, tuple) and len(size) == 3
        if any(s == 0 for s in size):
            return
        elif any(s < 0 for s in size):
            raise ValueError("Update size must not be negative")
        elif any(b < 0 for b in offset):
            raise ValueError("Update offset must not be negative")
        elif any(b + s > refsize for b, s, refsize in zip(offset, size, self.size)):
            raise ValueError("Update size out of range")
        # Apply - consider that texture arrays want to be uploaded per-texture
        # todo: avoid duplicates by merging with existing pending uploads
        if self.dim == 1:
            for z in range(size[2]):
                for y in range(size[1]):
                    offset2 = offset[0], y, z
                    size2 = size[0], 1, 1
                    self._pending_uploads.append((offset2, size2))
        elif self.dim == 2:
            for z in range(size[2]):
                offset2 = offset[0], offset[1], z
                size2 = size[0], size[1], 1
                self._pending_uploads.append((offset2, size2))
        else:
            self._pending_uploads.append((offset, size))
        self._rev += 1

    def _size_from_data(self, data, dim, size):
        # Check if shape matches dimension
        shape = data.shape

        if size:
            # Get version of size with trailing ones stripped
            size2 = size
            size2 = size2[:-1] if size2[-1] == 1 else size2
            size2 = size2[:-1] if size2[-1] == 1 else size2
            rsize = tuple(reversed(size2))
            # Check if size matches shape
            if rsize != shape[: len(rsize)]:
                raise ValueError(f"Given size does not match the data shape.")
            return size
        else:
            if len(shape) not in (dim, dim + 1):
                raise ValueError(
                    f"Can't map shape {shape} on {dim}D tex. Maybe also specify size?"
                )
            # Determine size based on dim and shape
            if dim == 1:
                return shape[0], 1, 1
            elif dim == 2:
                return shape[1], shape[0], 1
            else:  # dim == 3:
                return shape[2], shape[1], shape[0]

    def _get_subdata(self, offset, size, pixel_padding=None):
        """Return subdata as a contiguous array."""
        # If this is a full range, this is easy
        if offset == 0 and size == self.nitems and self.mem.contiguous:
            return self.mem
        # Get a numpy array, because memoryviews do not support nd slicing
        if isinstance(self.data, np.ndarray):
            arr = self.data
        elif not self.mem.c_contiguous:
            raise ValueError(
                "Non-contiguous texture data is only supported for numpy array."
            )
        else:
            arr = np.frombuffer(self.mem, self.mem.format)
        arr = arr.reshape(self.size[2], self.size[1], self.size[0], -1)
        # Slice it
        slices = []
        for d in reversed(range(3)):
            slices.append(slice(offset[d], offset[d] + size[d]))
        sub_arr = arr[tuple(slices)]
        if pixel_padding is not None:
            padding = np.ones(sub_arr.shape[:3] + (1,), dtype=sub_arr.dtype)
            sub_arr = np.concatenate([sub_arr, pixel_padding * padding], -1)
        return memoryview(np.ascontiguousarray(sub_arr))


def format_from_memoryview(mem, size):
    format = str(mem.format)
    format = STRUCT_FORMAT_ALIASES.get(format, format)
    # Process channels
    shape = mem.shape
    collapsed_size = [x for x in size if x > 1]
    if len(shape) == len(collapsed_size) + 1:
        nchannels = shape[-1]
    else:
        assert len(shape) == len(collapsed_size)
        nchannels = 1
    assert 1 <= nchannels <= 4
    tex_format = [None, "r", "rg", "rgb", "rgba"][nchannels]
    # if tex_format == "rgb":
    #     -> no raise: WGPU does not support rgb, but we handle it in the renderer
    # Process dtype. We select the tex_format that matches the dtype.
    # This means that uint8 values become 0..255 in the shader.
    # todo: not yet entirely sure about this
    # todo: there is no reference to wgpu here, but these are wgpu enums. Is that ok?
    texformatmap = {
        "b": "8snorm",
        "B": "8unorm",
        # "b": "8sint",
        # "B": "8uint",
        "h": "16sint",
        "H": "16uint",
        "i": "32sint",
        "U": "32uint",
        "e": "16float",
        "f": "32float",
    }
    if format in ("d", "float64"):
        raise TypeError("GPU's don't support float64 texture formats.")
    elif format not in texformatmap:
        raise TypeError(
            f"Cannot convert {format!r} to texture format. Maybe specify format?"
        )
    tex_format += texformatmap[format]
    return tex_format


# mipmaps: every texture can have a certain number of mipmap levels. Each
# next level is half the size of the previous level. I think we can design our
# API to target level 0 by default and allow a way to upload data to other levels.
# arrays: a d2_array view can be be created from a d2 texture with dept > 1
# cube: a special kind3 of array texture, with six 2D textures.
# cube_array: I suppose you'd have an array of 6xn textures in this case?


class TextureView(Resource):
    """A view on a texture.

    The view defines the sampling behavior and can specify a selection/different
    view on the texture.

    Passing no more than ``address_mode`` and ``filter`` will create a
    default view on the texture with the given sampling parameters.

    Parameters:
        address_mode (str): How to sample beyond the edges. Use "clamp",
            "mirror" or "repeat". Default "clamp".
            Can also use e.g. "clamp,clamp,repeat" to specify for u, v and w.
        filter (str): Interpolation filter. Use "nearest" or "linear".
            Default "nearest". Can also use e.g. "linear,linear,nearest" to set
            mag, min and mipmap filters.
        format (str): Omit or pass None to use the texture's format.
        view_dim (str): Omit or pass None to use the texture's format. Or e.g.
            get a "2d" slice view from a 3d texture, or e.g. "cube" or "2d-array".
        aspect (str): Omit or pass None to use the default.
        mip_range (range): A range object to specify what mip levels to view.
        layer_range (range): A range object to specify what array layers to view.
    """

    def __init__(
        self,
        texture,
        *,
        address_mode="clamp",
        filter="nearest",
        format=None,
        view_dim=None,
        aspect=None,
        mip_range=None,
        layer_range=None,
    ):
        self._rev = 1
        assert isinstance(texture, Texture)
        self._texture = texture
        # Sampler parameters
        self._address_mode = address_mode
        self._filter = filter
        # Texture view parameters
        self._format = format
        self._view_dim = view_dim
        self._aspect = aspect
        self._mip_range = mip_range or range(1)
        self._layer_range = layer_range or range(1)
        self._is_default_view = all(
            x is None for x in [format, view_dim, aspect, mip_range, layer_range]
        )

    @property
    def rev(self):
        # This is not actually increased anywhere, but it's added for consistency
        return self._rev

    @property
    def texture(self):
        """The Texture object holding the data for this texture view."""
        return self._texture

    @property
    def format(self):
        """The texture format."""
        return self._format or self.texture.format

    @property
    def view_dim(self):
        """The dimensionality of this view, as a string.
        See wgpu.TextureViewDimension.
        """
        return self._view_dim or f"{self.texture.dim}d"

    @property
    def address_mode(self):
        """How to sample beyond the edges. Use "clamp",
        "mirror" or "repeat". Default "clamp".
        """
        return self._address_mode

    @property
    def filter(self):
        """Interpolation filter. Use "nearest" or "linear"."""
        return self._filter



import math


class Spherical:
    _2pi = 2 * math.pi
    _eps = 0.000001

    def __init__(
        self, radius: float = 1.0, phi: float = 0.0, theta: float = 0.0
    ) -> None:
        self.radius = radius
        self.phi = phi
        self.theta = theta

    def set_from_vector3(self, v: "Vector3") -> "Spherical":
        return self.set_from_cartesian_coords(v.x, v.y, v.z)

    def set_from_cartesian_coords(self, x: float, y: float, z: float) -> "Spherical":
        self.radius = math.sqrt(x ** 2 + y ** 2 + z ** 2)
        if self.radius == 0:
            self.theta, self.phi = 0, 0
        else:
            self.theta = math.atan2(x, z)
            y /= self.radius
            if y < -1:
                y = -1
            elif y > 1:
                y = 1
            self.phi = math.acos(y)
        return self

    def make_safe(self) -> "Spherical":
        # restrict phi to (epsilon, pi-epsilon)
        self.phi = max(self._eps, min(math.pi - self._eps, self.phi))
        # restrict theta to (0, 2*pi)
        self.theta = self.theta % self._2pi
        return self



import numpy as np

from ._base import WorldObject
from ..resources import Buffer


class Mesh(WorldObject):
    """An object consisting of triangular faces, represented by vertices
    (3D positions) and an index that defines the connectivity.

    The picking info of a Mesh (the result of
    ``renderer.get_pick_info()``) will for most materials include
    ``instance_index`` (int), ``face_index`` (int), and ``face_coords``
    (tuple of 3 floats). The latter are the barycentric coordinates for
    each vertex of the face (with values 0..1).


    """

    def __init__(self, geometry, material):
        super().__init__()
        self.geometry = geometry
        self.material = material


class InstancedMesh(Mesh):
    def __init__(self, geometry, material, count):
        super().__init__(geometry, material)
        count = int(count)
        # Create count eye matrices
        matrices = np.zeros((count, 4, 4), np.float32)
        for i in range(4):
            matrices[:, i, i] = 1
        self.matrices = Buffer(matrices, usage="STORAGE", nitems=count)

    def set_matrix_at(self, index: int, matrix):
        matrix = np.array(matrix).reshape(4, 4)
        self.matrices.data[index] = matrix



from ._base import WorldObject


class Group(WorldObject):
    """A Group is almost identical to a WorldObject. Its purpose is
    to make working with groups of objects syntactically clearer.
    """



import wgpu  # only for flags/enums

from . import register_wgpu_render_function
from ._shadercomposer import BaseShader
from ...objects import Volume
from ...materials import VolumeSliceMaterial
from ...resources import Texture, TextureView


@register_wgpu_render_function(Volume, VolumeSliceMaterial)
def volume_slice_renderer(wobject, render_info):
    """Render function capable of rendering volumes."""

    geometry = wobject.geometry
    material = wobject.material  # noqa
    shader = VolumeSliceShader(climcorrection=False)

    shader.define_uniform(0, 0, "u_stdinfo", render_info.stdinfo_uniform.data.dtype)
    shader.define_uniform(0, 1, "u_wobject", wobject.uniform_buffer.data.dtype)
    shader.define_uniform(0, 2, "u_material", material.uniform_buffer.data.dtype)

    bindings0 = {
        0: ("buffer/uniform", render_info.stdinfo_uniform),
        1: ("buffer/uniform", wobject.uniform_buffer),
        2: ("buffer/uniform", material.uniform_buffer),
    }

    topology = wgpu.PrimitiveTopology.triangle_list
    n = 12

    # Collect texture and sampler
    if material.map is None:
        raise ValueError("VolumeSliceMaterial must have a texture map.")
    else:
        if isinstance(material.map, TextureView):
            view = material.map
        elif isinstance(material.map, Texture):
            view = material.map.get_view(filter="linear")
        else:
            raise TypeError("material.map must be a TextureView")
        if view.view_dim.lower() != "3d":
            raise TypeError("material.map must a 3D texture (view)")
        elif getattr(geometry, "texcoords", None) is None:
            raise ValueError("With VolumeSliceMaterial, geometry needs texcoords")
        # Sampling type
        if "norm" in material.map.format or "float" in material.map.format:
            shader["texture_format"] = "f32"
            if "unorm" in material.map.format:
                shader["climcorrection"] = " * 255.0"
            elif "snorm" in material.map.format:
                shader["climcorrection"] = " * 255.0 - 128.0"
        elif "uint" in material.map.format:
            shader["texture_format"] = "u32"
        else:
            shader["texture_format"] = "i32"
        # Channels
        if material.map.format.startswith("rgb"):  # rgb maps to rgba
            shader["texture_color"] = True
        elif material.map.format.startswith("r"):
            shader["texture_color"] = False
        else:
            raise ValueError("Unexpected texture format")

    bindings1 = {
        0: ("sampler/filtering", view),
        1: ("texture/auto", view),
        2: ("buffer/read_only_storage", geometry.positions),
        3: ("buffer/read_only_storage", geometry.texcoords),
    }

    # Put it together!
    wgsl = shader.generate_wgsl()
    return [
        {
            "vertex_shader": (wgsl, "vs_main"),
            "fragment_shader": (wgsl, "fs_main"),
            "primitive_topology": topology,
            "indices": (range(n), range(1)),
            "vertex_buffers": {},
            "bindings0": bindings0,
            "bindings1": bindings1,
        }
    ]


class VolumeSliceShader(BaseShader):
    def get_code(self):
        return (
            self.get_definitions()
            + self.more_definitions()
            + self.vertex_shader()
            + self.fragment_shader()
        )

    def more_definitions(self):
        return """

        struct VertexInput {
            [[builtin(vertex_index)]] index : u32;
        };
        struct VertexOutput {
            [[location(0)]] texcoord: vec3<f32>;
            [[builtin(position)]] pos: vec4<f32>;
        };

        struct FragmentOutput {
            [[location(0)]] color: vec4<f32>;
            [[location(1)]] pick: vec4<i32>;
        };

        [[block]]
        struct BufferF32 {
            data: [[stride(4)]] array<f32>;
        };


        [[group(1), binding(0)]]
        var r_sampler: sampler;

        [[group(1), binding(1)]]
        var r_tex: texture_3d<{{ texture_format }}>;

        [[group(1), binding(2)]]
        var<storage> s_positions: [[access(read)]] BufferF32;

        [[group(1), binding(3)]]
        var<storage> s_texcoords: [[access(read)]] BufferF32;
        """

    def vertex_shader(self):
        return """

        [[stage(vertex)]]
        fn vs_main(in: VertexInput) -> VertexOutput {
            var out: VertexOutput;

            // We're assuming a box geometry, using the same layout as a simple
            // ThreeJS BoxBufferGeometry. And we're only using the first eight
            // vertices. These are laid out like this:
            //
            //   Vertices       Planes (right, left, back, front, top, bottom)
            //                            0      1    2      3     4     5
            //
            //    5----0        0: 0231        +----+
            //   /|   /|        1: 7546       /|24 /|
            //  7----2 |        2: 5014      +----+ |0
            //  | 4--|-1        3: 2763     1| +--|-+
            //  |/   |/         4: 0572      |/35 |/
            //  6----3          5: 3641      +----+


            let plane = u_material.plane.xyzw;  // ax + by + cz + d
            let n = plane.xyz;

            // Define edges (using vertex indices), and their matching plane
            // indices (each edge touches two planes). Note that these need to
            // match the above figure, and that needs to match with the actual
            // BoxGeometry implementation!
            let edges = array<vec2<i32>,12>(
                vec2<i32>(0, 2), vec2<i32>(2, 3), vec2<i32>(3, 1), vec2<i32>(1, 0),
                vec2<i32>(4, 6), vec2<i32>(6, 7), vec2<i32>(7, 5), vec2<i32>(5, 4),
                vec2<i32>(5, 0), vec2<i32>(1, 4), vec2<i32>(2, 7), vec2<i32>(6, 3),
            );
            let ed2pl = array<vec2<i32>,12>(
                vec2<i32>(0, 4), vec2<i32>(0, 3), vec2<i32>(0, 5), vec2<i32>(0, 2),
                vec2<i32>(1, 5), vec2<i32>(1, 3), vec2<i32>(1, 4), vec2<i32>(1, 2),
                vec2<i32>(2, 4), vec2<i32>(2, 5), vec2<i32>(3, 4), vec2<i32>(3, 5),
            );

            // Init intersection info
            var intersect_flags: array<i32,12> = array<i32,12>(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
            var intersect_positions: array<vec3<f32>,12> = array<vec3<f32>,12>(
                vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0),
            );
            var intersect_texcoords: array<vec3<f32>,12> = array<vec3<f32>,12>(
                vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0), vec3<f32>(0.0, 0.0, 0.0),
            );

            // Intersect the 12 edges
            for (var i:i32=0; i<12; i=i+1) {
                let edge = edges[i];
                let p1_raw = vec3<f32>(
                    s_positions.data[edge[0] * 3],
                    s_positions.data[edge[0] * 3 + 1],
                    s_positions.data[edge[0] * 3 + 2],
                );
                let p2_raw = vec3<f32>(
                    s_positions.data[edge[1] * 3],
                    s_positions.data[edge[1] * 3 + 1],
                    s_positions.data[edge[1] * 3 + 2],
                );
                let p1_ = u_wobject.world_transform * vec4<f32>(p1_raw, 1.0);
                let p2_ = u_wobject.world_transform * vec4<f32>(p2_raw, 1.0);
                let p1 = p1_.xyz / p1_.w;
                let p2 = p2_.xyz / p2_.w;
                let tc1 = vec3<f32>(
                    s_texcoords.data[edge[0] * 3],
                    s_texcoords.data[edge[0] * 3 + 1],
                    s_texcoords.data[edge[0] * 3 + 2],
                );
                let tc2 = vec3<f32>(
                    s_texcoords.data[edge[1] * 3],
                    s_texcoords.data[edge[1] * 3 + 1],
                    s_texcoords.data[edge[1] * 3 + 2],
                );
                let u = p2 - p1;
                let t = -(plane.x * p1.x + plane.y * p1.y + plane.z * p1.z + plane.w) / dot(n, u);
                let intersects:bool = t > 0.0 && t < 1.0;
                intersect_flags[i] = select(1, 0, intersects);
                intersect_positions[i] = mix(p1, p2, vec3<f32>(t, t, t));
                intersect_texcoords[i] = mix(tc1, tc2, vec3<f32>(t, t, t));
            }

            // Init six vertices
            var vertices: array<vec3<f32>,6> = array<vec3<f32>,6>(
                vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0),
            );
            var texcoords: array<vec3<f32>,6> = array<vec3<f32>,6>(
                vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0),
                vec3<f32>(0.0, 0.0, 0.0),
            );

            // Find first intersection point. This can be any valid intersection.
            // In ed2pl[i][0], the 0 could also be a one. It would mean that we'd
            // move around the box in the other direction.
            var plane_index: i32 = 0;
            for (var i:i32=0; i<12; i=i+1) {
                if (intersect_flags[i] == 1) {
                    plane_index = ed2pl[i][0];
                    vertices[0] = intersect_positions[i];
                    texcoords[0] = intersect_texcoords[i];
                    break;
                }
            }

            // From there take (at most) 5 steps
            let i_start: i32 = i;
            var i_last: i32 = i;
            var max_iter: i32 = 6;
            for (var iter:i32=1; iter<max_iter; iter=iter+1) {
                for (var i:i32=0; i<12; i=i+1) {
                    if (i != i_last && intersect_flags[i] == 1) {
                        if (ed2pl[i][0] == plane_index) {
                            vertices[iter] = intersect_positions[i];
                            texcoords[iter] = intersect_texcoords[i];
                            plane_index = ed2pl[i][1];
                            i_last = i;
                            break;
                        } elseif (ed2pl[i][1] == plane_index) {
                            vertices[iter] = intersect_positions[i];
                            texcoords[iter] = intersect_texcoords[i];
                            plane_index = ed2pl[i][0];
                            i_last = i;
                            break;
                        }
                    }
                }
                if (i_last == i_start) {
                    max_iter = iter;
                    break;
                }
            }

            // Make the rest degenerate triangles
            for (var i:i32=max_iter; i<6; i=i+1) {
                vertices[i] = vertices[0];
            }

            // Now select the current vertex. We mimic a triangle fan with a triangle list.
            // This works whether the number of vertices/intersections is 3, 4, 5, and 6.
            let index = i32(in.index);
            let indexmap = array<i32,12>(0, 1, 2, 0, 2, 3, 0, 3, 4, 0, 4, 5);
            let world_pos = vertices[ indexmap[index] ];
            let ndc_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * vec4<f32>(world_pos, 1.0);
            out.pos = ndc_pos;
            out.texcoord = texcoords[ indexmap[index] ];

            return out;
        }
        """

    def fragment_shader(self):
        return """

        [[stage(fragment)]]
        fn fs_main(in: VertexOutput) -> FragmentOutput {
            var out: FragmentOutput;
            var color_value: vec4<f32>;

            $$ if texture_format == 'f32'
                color_value = textureSample(r_tex, r_sampler, in.texcoord.xyz);
            $$ else
                let texcoords_dim = vec3<f32>(textureDimensions(r_tex));
                let texcoords_u = vec3<i32>(in.texcoord.xyz * texcoords_dim % texcoords_dim);
                color_value = vec4<f32>(textureLoad(r_tex, texcoords_u, 0));
            $$ endif

            $$ if climcorrection
                color_value = vec4<f32>(color_value.rgb {{ climcorrection }}, color_value.a);
            $$ endif
            $$ if not texture_color
                color_value = color_value.rrra;
            $$ endif
            let albeido = (color_value.rgb - u_material.clim[0]) / (u_material.clim[1] - u_material.clim[0]);

            out.color = vec4<f32>(albeido, color_value.a);
            out.pick = vec4<i32>(u_wobject.id, vec3<i32>(in.texcoord * 1048576.0 + 0.5));
            return out;
        }
        """



from ..objects import WorldObject
from ..materials import Material


class Renderer:
    """Base class for other renderers. A renderer takes a figure,
    collect data that describe how it should be drawn, and then draws it.
    """

    pass


class RenderFunctionRegistry:
    """A registry for render functions capable of rendering specific
    object-material combinations. This registry allows for a scalable
    plugin-like system for a renderer's capabilities.
    """

    def __init__(self):
        self._store = {}
        self._known_classes = set([WorldObject, Material])

    def register(self, wobject_cls, material_cls, func):
        """Register a render function for the given combination of
        world object and material class.

        When a world object is being rendered, the renderer will select
        a render function based on the types of the world object and
        its material. The selection prioritizes more specialized
        classes. Therefore, custom render functions should be registered
        as specifically as possible to avoid clashes. Builtin functions
        can register "catch all" functions, e.g. providing a solution
        for all materials with a specific world object subclass, using
        with (SomeWorldObject, Material).
        """

        # Check types
        if not (isinstance(wobject_cls, type) and issubclass(wobject_cls, WorldObject)):
            raise TypeError(f"Expected WorldObject subclass, got {wobject_cls}.")
        if not (isinstance(material_cls, type) and issubclass(material_cls, Material)):
            raise TypeError(f"Expected Material subclass, got {material_cls}.")
        if not callable(func):
            raise TypeError("Third argument must be a callable, not {func}")

        # Store func with a key derived from the classes
        key = (wobject_cls, material_cls)
        if key in self._store:
            raise ValueError(
                "Each combination of WorldObject and Material can only be registered once."
            )
        self._store[key] = func
        self._known_classes.add(wobject_cls)
        self._known_classes.add(material_cls)

    def get_render_function(self, wobject):
        """Get the render function for the given world object, based
        on the object's class, and the class of its material. The behavior
        is similar to ``isinstance``; providing an instance of a custom
        subclasses should not affect the selection. Returns None if the
        object has no material, or a suitable renderer cannot be
        selected.
        """

        # Get and check world object type
        wobject_cls = type(wobject)
        if not isinstance(wobject, WorldObject):
            raise TypeError(f"Expected WorldObject instance, got {wobject_cls}.")

        # Get the material, if the object has it
        material = getattr(wobject, "material", None)
        if material is None:
            return None

        # Get and check the material type
        material_cls = type(material)
        if not isinstance(material, Material):
            raise TypeError(f"Expected Material instance, got {material_cls}.")

        # Get WorldObject mro, but stripped to classes of interest
        wobject_classes = wobject_cls.mro()
        while wobject_classes[-1] is not WorldObject:
            wobject_classes.pop(-1)
        while wobject_classes[0] not in self._known_classes:
            wobject_classes.pop(0)

        # Get Material mro, but stripped to classes of interest
        material_classes = material_cls.mro()
        while material_classes[-1] is not Material:
            material_classes.pop(-1)
        while material_classes[0] not in self._known_classes:
            material_classes.pop(0)

        # Try to select a renderer. There is a double-loop here, but
        # both iters should be short, especially with the stripping
        # above. An alternative implementation would be to walk over
        # all registry entries and use isinstance, but the number of
        # render functions will (over time) exceed the depth of
        # subclasses that is traversed here.
        for cls1 in wobject_classes:
            for cls2 in material_classes:
                key = cls1, cls2
                f = self._store.get(key, None)
                if f is not None:
                    return f

        return None



import gc
import random
import weakref

from ..linalg import Vector3, Matrix4, Quaternion
from ..resources import Resource, Buffer
from ..utils import array_from_shadertype


# Keep track of id's. About the max:
# * 2_147_483_647 (2**31 -1) max number for signed i32.
# *    16_777_216 max integer that can be stored exactly in f32
# *     4_000_000 max integer that survives being passed as a varying (in my tests)
_idmap = weakref.WeakKeyDictionary()
_idmax = 16_777_217  # non-inclusive


class ResourceContainer:
    """Base class for WorldObject, Geometry and Material."""

    def __init__(self):
        self._resource_parents = weakref.WeakSet()
        self._rev = 0

    @property
    def rev(self):
        """Monotonically increasing integer that gets bumped when any
        of its buffers or textures are set. (Not when updates are made
        to these resources themselves).
        """
        return self._rev

    # NOTE: we could similarly let bumping of a resource's rev bump a
    # data_rev here. But it is not clear whether the (minor?) increase
    # in performance is worth the added complexity.

    def _bump_rev(self):
        """Bump the rev (and that of any "parents")"""
        self._rev += 1
        for x in self._resource_parents:
            x._rev += 1

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if isinstance(value, ResourceContainer):
            value._resource_parents.add(self)
            self._bump_rev()
        elif isinstance(value, Resource):
            self._bump_rev()


class WorldObject(ResourceContainer):
    """The base class for objects present in the "world", i.e. the scene graph.

    Each WorldObject has geometry to define it's data, and material to define
    its apearance. The object itself is only responsible for defining object
    hierarchies (parent / children) and its position and orientation in the world.

    This is considered a base class. Use Group to collect multiple world objects
    into a single empty world object.
    """

    # The uniform type describes the structured info for this object, which represents
    # every "propery" that a renderer would need to know in order to visualize it.
    # Put larger items first for alignment, also note that host-sharable structs
    # align at power-of-two only, so e.g. vec3 needs padding.
    # todo: rename uniform to info or something?

    uniform_type = dict(
        world_transform=("float32", (4, 4)),
        id=("int32",),
    )

    _v = Vector3()
    _m = Matrix4()
    _q = Quaternion()

    def __init__(self):
        super().__init__()
        self.parent = None
        self._children = []

        self.position = Vector3()
        self.rotation = Quaternion()
        self.scale = Vector3(1, 1, 1)
        self._transform_hash = ()

        self.up = Vector3(0, 1, 0)

        self.matrix = Matrix4()
        self.matrix_auto_update = True
        self.matrix_world = Matrix4()
        self.matrix_world_dirty = True

        self.uniform_buffer = Buffer(
            array_from_shadertype(self.uniform_type), usage="uniform"
        )

        self.visible = True
        self.render_order = 0

        # See if we reached max number of objects. If so, try cleanup and try again.
        # Actually, stop earlier to prevent that while loop below to become slow.
        max_items = 0.75 * _idmax
        if len(_idmap) >= max_items:
            gc.collect()
            if len(_idmap) >= max_items:
                raise RuntimeError("Max number of objects reached")
        # Set id
        self._id = random.randint(1, _idmax - 1)
        while self._id in _idmap.values():
            self._id = (self._id + 1) % _idmax
        _idmap[self] = self._id
        self.uniform_buffer.data["id"] = self._id

    @property
    def id(self):
        """An integer id smaller than 2**31."""
        return self._id

    @property
    def children(self):
        return tuple(self._children)

    def add(self, obj):
        # orphan if needed
        if obj.parent is not None:
            obj.parent.remove(obj)
        # attach to scene graph
        obj.parent = self
        self._children.append(obj)
        # flag world matrix as dirty
        obj.matrix_world_dirty = True
        return self

    def remove(self, obj):
        if obj in self._children:
            obj.parent = None
            self._children.remove(obj)

    def traverse(self, callback):
        callback(self)
        for child in self.children:
            child.traverse(callback)

    def update_matrix(self):
        p, r, s = self.position, self.rotation, self.scale
        hash = p.x, p.y, p.z, r.x, r.y, r.z, r.w, s.x, s.y, s.z
        if hash != self._transform_hash:
            self._transform_hash = hash
            self.matrix.compose(self.position, self.rotation, self.scale)
            self.matrix_world_dirty = True

    def set_matrix(self, matrix):
        self.matrix.copy(matrix)
        self.matrix.decompose(self.position, self.rotation, self.scale)
        self.matrix_world_dirty = True

    def apply_matrix(self, matrix):
        if self.matrix_auto_update:
            self.update_matrix()
        self.matrix.premultiply(matrix)
        self.matrix.decompose(self.position, self.rotation, self.scale)
        self.matrix_world_dirty = True

    def update_matrix_world(
        self, force=False, update_children=True, update_parents=False
    ):
        if update_parents and self.parent:
            self.parent.update_matrix_world(
                force=force, update_children=False, update_parents=True
            )
        if self.matrix_auto_update:
            self.update_matrix()
        if self.matrix_world_dirty or force:
            if self.parent is None:
                self.matrix_world.copy(self.matrix)
            else:
                self.matrix_world.multiply_matrices(
                    self.parent.matrix_world, self.matrix
                )
            self.uniform_buffer.data[
                "world_transform"
            ].flat = self.matrix_world.elements
            self.uniform_buffer.update_range(0, 1)
            self.matrix_world_dirty = False
            for child in self._children:
                child.matrix_world_dirty = True
        if update_children:
            for child in self._children:
                child.update_matrix_world()

    def look_at(self, target: Vector3):
        self.update_matrix_world(update_parents=True, update_children=False)
        self._v.set_from_matrix_position(self.matrix_world)
        self._m.look_at(self._v, target, self.up)
        self.rotation.set_from_rotation_matrix(self._m)
        if self.parent:
            self._m.extract_rotation(self.parent.matrix_world)
            self._q.set_from_rotation_matrix(self._m)
            self.rotation.premultiply(self._q.inverse())



from . import register_svg_render_function
from ...objects import Line
from ...materials import LineMaterial


@register_svg_render_function(Line, LineMaterial)
def line_renderer(wobject):
    """Render function capable of rendering lines."""

    geometry = wobject.geometry
    positions = geometry.positions.data

    style = "stroke:rgb(255,0,0);stroke-width:2"

    lines = []
    for i in range(len(positions) - 1):
        x1 = positions[i, 0]
        y1 = positions[i, 1]
        x2 = positions[i + 1, 0]
        y2 = positions[i + 1, 1]
        lines.append(
            f"<line x1='{x1}' y1='{y1}' x2='{x2}' y2='{y2}' style='{style}' />"
        )

    return lines



from math import cos, sin, acos, atan2

from .utils import clamp, MACHINE_EPSILON


class Quaternion:
    def __init__(self, x: float = 0, y: float = 0, z: float = 0, w: float = 1) -> None:
        self.x = x
        self.y = y
        self.z = z
        self.w = w

    def set(self, x: float, y: float, z: float, w: float) -> "Quaternion":
        self.x = x
        self.y = y
        self.z = z
        self.w = w
        return self

    def clone(self) -> "Quaternion":
        return Quaternion(self.x, self.y, self.z, self.w)

    def __repr__(self) -> str:
        return f"Quaternion({self.x}, {self.y}, {self.z}, {self.w})"

    def copy(self, quaternion: "Quaternion") -> "Quaternion":
        self.x = quaternion.x
        self.y = quaternion.y
        self.z = quaternion.z
        self.w = quaternion.w
        return self

    def set_from_euler(self, euler: "Euler") -> "Quaternion":
        from .euler import Euler

        x = euler.x
        y = euler.y
        z = euler.z
        order = euler.order

        c1 = cos(x / 2)
        c2 = cos(y / 2)
        c3 = cos(z / 2)
        s1 = sin(x / 2)
        s2 = sin(y / 2)
        s3 = sin(z / 2)

        if order == Euler.RotationOrders.XYZ:
            self.x = s1 * c2 * c3 + c1 * s2 * s3
            self.y = c1 * s2 * c3 - s1 * c2 * s3
            self.z = c1 * c2 * s3 + s1 * s2 * c3
            self.w = c1 * c2 * c3 - s1 * s2 * s3
        elif order == Euler.RotationOrders.YXZ:
            self.x = s1 * c2 * c3 + c1 * s2 * s3
            self.y = c1 * s2 * c3 - s1 * c2 * s3
            self.z = c1 * c2 * s3 - s1 * s2 * c3
            self.w = c1 * c2 * c3 + s1 * s2 * s3
        elif order == Euler.RotationOrders.ZXY:
            self.x = s1 * c2 * c3 - c1 * s2 * s3
            self.y = c1 * s2 * c3 + s1 * c2 * s3
            self.z = c1 * c2 * s3 + s1 * s2 * c3
            self.w = c1 * c2 * c3 - s1 * s2 * s3
        elif order == Euler.RotationOrders.ZYX:
            self.x = s1 * c2 * c3 - c1 * s2 * s3
            self.y = c1 * s2 * c3 + s1 * c2 * s3
            self.z = c1 * c2 * s3 - s1 * s2 * c3
            self.w = c1 * c2 * c3 + s1 * s2 * s3
        elif order == Euler.RotationOrders.YZX:
            self.x = s1 * c2 * c3 + c1 * s2 * s3
            self.y = c1 * s2 * c3 + s1 * c2 * s3
            self.z = c1 * c2 * s3 - s1 * s2 * c3
            self.w = c1 * c2 * c3 - s1 * s2 * s3
        elif order == Euler.RotationOrders.XZY:
            self.x = s1 * c2 * c3 - c1 * s2 * s3
            self.y = c1 * s2 * c3 - s1 * c2 * s3
            self.z = c1 * c2 * s3 + s1 * s2 * c3
            self.w = c1 * c2 * c3 + s1 * s2 * s3

        return self

    def set_from_axis_angle(self, axis: "Vector3", angle: float) -> "Quaternion":
        # assumes axis is normalized
        half_angle = angle / 2
        s = sin(half_angle)

        self.x = axis.x * s
        self.y = axis.y * s
        self.z = axis.z * s
        self.w = cos(half_angle)

        return self

    def set_from_rotation_matrix(self, m: "Matrix4") -> "Quaternion":
        # assumes the upper 3x3 of m is a pure rotation matrix (i.e, unscaled)
        te = m.elements
        m11 = te[0]
        m12 = te[4]
        m13 = te[8]
        m21 = te[1]
        m22 = te[5]
        m23 = te[9]
        m31 = te[2]
        m32 = te[6]
        m33 = te[10]
        trace = m11 + m22 + m33

        if trace > 0:
            s = 0.5 / ((trace + 1.0) ** 0.5)
            self.w = 0.25 / s
            self.x = (m32 - m23) * s
            self.y = (m13 - m31) * s
            self.z = (m21 - m12) * s

        elif m11 > m22 and m11 > m33:
            s = 2.0 * ((1.0 + m11 - m22 - m33) ** 0.5)
            self.w = (m32 - m23) / s
            self.x = 0.25 * s
            self.y = (m12 + m21) / s
            self.z = (m13 + m31) / s

        elif m22 > m33:
            s = 2.0 * ((1.0 + m22 - m11 - m33) ** 0.5)
            self.w = (m13 - m31) / s
            self.x = (m12 + m21) / s
            self.y = 0.25 * s
            self.z = (m23 + m32) / s

        else:
            s = 2.0 * ((1.0 + m33 - m11 - m22) ** 0.5)
            self.w = (m21 - m12) / s
            self.x = (m13 + m31) / s
            self.y = (m23 + m32) / s
            self.z = 0.25 * s

        return self

    def set_from_unit_vectors(self, v_from: "Vector3", v_to: "Vector3") -> "Quaternion":
        # assumes direction vectors v_from and v_to are normalized
        eps = 0.000001
        r = v_from.dot(v_to) + 1

        if r < eps:
            r = 0
            if abs(v_from.x) > abs(v_from.z):
                self.x = -v_from.y
                self.y = v_from.x
                self.z = 0
                self.w = r
            else:
                self.x = 0
                self.y = -v_from.z
                self.z = v_from.y
                self.w = r
        else:
            self.x = v_from.y * v_to.z - v_from.z * v_to.y
            self.y = v_from.z * v_to.x - v_from.x * v_to.z
            self.z = v_from.x * v_to.y - v_from.y * v_to.x
            self.w = r

        return self.normalize()

    def angle_to(self, q: "Quaternion") -> float:
        return 2 * acos(abs(clamp(self.dot(q), -1, 1)))

    def rotate_towards(self, q: "Quaternion", step: float) -> "Quaternion":
        angle = self.angle_to(q)
        if angle == 0:
            return self
        t = min(1, step / angle)
        self.slerp(q, t)
        return self

    def inverse(self) -> "Quaternion":
        # quaternion is assumed to have unit length
        return self.conjugate()

    def conjugate(self) -> "Quaternion":
        self.x *= -1
        self.y *= -1
        self.z *= -1
        return self

    def dot(self, q: "Quaternion") -> float:
        return self.x * q.x + self.y * q.y + self.z * q.z + self.w * q.w

    def length_sq(self) -> float:
        return self.x * self.x + self.y * self.y + self.z * self.z + self.w * self.w

    def length(self) -> float:
        return self.length_sq() ** 0.5

    def normalize(self) -> "Quaternion":
        length = self.length()

        if length == 0:
            self.x = 0
            self.y = 0
            self.z = 0
            self.w = 1
        else:
            self.x /= length
            self.y /= length
            self.z /= length
            self.w /= length

        return self

    def multiply(self, q: "Quaternion") -> "Quaternion":
        return self.multiply_quaternions(self, q)

    def premultiply(self, q: "Quaternion") -> "Quaternion":
        return self.multiply_quaternions(q, self)

    def multiply_quaternions(self, a: "Quaternion", b: "Quaternion") -> "Quaternion":
        qax = a.x
        qay = a.y
        qaz = a.z
        qaw = a.w
        qbx = b.x
        qby = b.y
        qbz = b.z
        qbw = b.w

        self.x = qax * qbw + qaw * qbx + qay * qbz - qaz * qby
        self.y = qay * qbw + qaw * qby + qaz * qbx - qax * qbz
        self.z = qaz * qbw + qaw * qbz + qax * qby - qay * qbx
        self.w = qaw * qbw - qax * qbx - qay * qby - qaz * qbz

        return self

    def slerp(self, qb: "Quaternion", t: float) -> "Quaternion":
        if t == 0:
            return self
        elif t == 1:
            return self.copy(qb)

        x = self.x
        y = self.y
        z = self.z
        w = self.w
        cos_half_theta = w * qb.w + x * qb.x + y * qb.y + z * qb.z
        if cos_half_theta < 0:
            self.w = -qb.w
            self.x = -qb.x
            self.y = -qb.y
            self.z = -qb.z
            cos_half_theta = -cos_half_theta
        else:
            self.copy(qb)

        if cos_half_theta >= 1.0:
            self.w = w
            self.x = x
            self.y = y
            self.z = z
            return self

        sqr_sin_half_theta = 1.0 - cos_half_theta * cos_half_theta
        if sqr_sin_half_theta <= MACHINE_EPSILON:
            s = 1 - t
            self.w = s * w + t * self.w
            self.x = s * x + t * self.x
            self.y = s * y + t * self.y
            self.z = s * z + t * self.z
            self.normalize()
            return self

        sin_half_theta = sqr_sin_half_theta ** 0.5
        half_theta = atan2(sin_half_theta, cos_half_theta)
        ratio_a = sin((1 - t) * half_theta) / sin_half_theta
        ratio_b = sin(t * half_theta) / sin_half_theta

        self.w = w * ratio_a + self.w * ratio_b
        self.x = x * ratio_a + self.x * ratio_b
        self.y = y * ratio_a + self.y * ratio_b
        self.z = z * ratio_a + self.z * ratio_b
        return self

    def equals(self, quaternion: "Quaternion") -> bool:
        return (
            quaternion.x == self.x
            and quaternion.y == self.y
            and quaternion.z == self.z
            and quaternion.w == self.w
        )

    def __eq__(self, other: "Quaternion") -> bool:
        return isinstance(other, Quaternion) and self.equals(other)

    def from_array(self, array: list, offset: int = 0) -> "Quaternion":
        self.x = array[offset]
        self.y = array[offset + 1]
        self.z = array[offset + 2]
        self.w = array[offset + 3]
        return self

    def to_array(self, array: list = None, offset: int = 0) -> list:
        if array is None:
            array = []

        padding = offset + 4 - len(array)
        if padding > 0:
            array.extend((None for _ in range(padding)))

        array[offset] = self.x
        array[offset + 1] = self.y
        array[offset + 2] = self.z
        array[offset + 3] = self.w
        return array



from ._base import WorldObject


class Points(WorldObject):
    """An object consisting of points represented by vertices (3D positions).

    The picking info of a Points object (the result of
    ``renderer.get_pick_info()``) will for most materials include
    ``vertex_index`` (int).

    """

    def __init__(self, geometry, material):
        super().__init__()
        self.geometry = geometry
        self.material = material



import numpy as np

from ..resources import Buffer
from ._base import Geometry


class KleinBottleGeometry(Geometry):
    """The Klein bottle is a surface for which the inside and outside
    are the same. A bit like a Möbius strip. In fact, a Klein bottle
    can be constructed by glueing together two Möbius strips.

    More technically: the Klein bottle is an example of a non-orientable
    surface; it is a two-dimensional manifold against which a system
    for determining a normal vector cannot be consistently defined.
    """

    # This is an interesting object for mathematicians. For us it's
    # interesting because we can test whether our lighting etc. deals
    # correctly with objects for which the "inside" must also be shown.

    def __init__(self, scale=1.0):
        super().__init__()

        # The number of vertices is nxn
        n = 40

        # Get 2D surface in 3D space
        u = np.linspace(0, 2 * np.pi, n, endpoint=True, dtype=np.float32)
        v = np.linspace(0, 2 * np.pi, n, endpoint=False, dtype=np.float32)
        ux, vx = np.meshgrid(u, v)
        x, y, z = klein_bottle_surface(ux, vx)

        # Scaled to a unit cube, then scale to width / height / depth
        # x = (x + 1.66559) * (0.0437597 * width)
        # y = (y - 2.04939) * (0.0277017 * height)
        # z = (z + 0.00000) * (0.0833333 * depth)
        # Scaled to fit inside a unit cube, but maintain original proportions
        x = (x + 1.66559) * (0.0833333 * scale)
        y = (y - 2.04939) * (0.0833333 * scale)
        z = (z + 0.00000) * (0.0833333 * scale)

        # Put into an Nx4 array
        positions = np.empty((x.size, 3), np.float32)
        positions[:, 0] = x.flat
        positions[:, 1] = y.flat
        positions[:, 2] = z.flat

        # Texcoords are easy
        texcoords = np.column_stack([ux.flat, vx.flat]).astype(np.float32, copy=False)
        texcoords *= 1 / (2 * np.pi)

        # Map indices
        # Two triangles onto the "top-left" rectangle (six vertices)
        indices = np.array([0, 1, n + 1, n + 1, n, 0], np.int32)
        # Replicate to all rectangles, add offsets
        indices = np.tile(indices, (n, n - 1, 1))
        gx, gy = np.meshgrid(
            np.arange(indices.shape[1]), n * np.arange(indices.shape[0])
        )
        indices += (gx + gy).reshape(indices.shape[:2] + (1,))
        # Stitch the ends together over one axis. We can't stitch the other ends
        # together, since that's where the normals flip from "inside" to "outside".
        indices[-1, :, 2:5] -= n * n
        indices = indices.reshape(-1)

        # Create buffers for this geometry
        self.positions = Buffer(positions, usage="vertex|storage")
        self.index = Buffer(indices, usage="index|storage")
        self.texcoords = Buffer(texcoords, usage="vertex|storage")


def klein_bottle_surface(u, v):
    """
    http://paulbourke.net/geometry/toroidal/

        A German topologist named Klein
        Thought the Möbius Loop was divine
        Said he, "If you glue
        The edges of two
        You get a weird bottle like mine.
    """
    half = (0 <= u) & (u < np.pi)
    r = 4 * (1 - np.cos(u) / 2)
    x = 6 * np.cos(u) * (1 + np.sin(u)) + r * np.cos(v + np.pi)
    x[half] = (6 * np.cos(u) * (1 + np.sin(u)) + r * np.cos(u) * np.cos(v))[half]
    y = 16 * np.sin(u)
    y[half] = (16 * np.sin(u) + r * np.sin(u) * np.cos(v))[half]
    z = r * np.sin(v)
    return x, y, z


class TorusKnotGeometry(Geometry):
    """Creates a torus knot, the particular shape of which is defined
    by a pair of coprime integers, p and q. If p and q are not coprime,
    the result will be a torus link.

    Arguments:
        scale (float): the scale of the torus, default 1.
        tube (float): the radius of the tube. Default 0.4.
        tubular_segments (int): default is 64.
        radial_segments (int): default is 8.
        p (int): how many times the geometry winds around its axis of
            rotational symmetry. Default 2.
        q (int): how many times the geometry winds around a circle in
            the interior of the torus. Default 3.

    """

    def __init__(
        self, scale=1.0, tube=0.4, tubular_segments=64, radial_segments=8, p=2, q=3
    ):
        super().__init__()

        # If we stitch, we have no duplicate vertices, but stitch the
        # ends together with the indices, resulting in a fully closed object.
        # However, texturing works better without such stitching.
        stitch = False

        if stitch:
            tubular_verts = tubular_segments
            radial_verts = radial_segments
        else:
            tubular_verts = tubular_segments + 1
            radial_verts = radial_segments + 1

        # Define base factors
        u = np.linspace(
            0, p * 2 * np.pi, tubular_verts, endpoint=not stitch, dtype=np.float32
        )
        v = np.linspace(
            0, 2 * np.pi, radial_verts, endpoint=not stitch, dtype=np.float32
        )

        # Get positions along the torus' center, and a tiny step further
        pos1 = torus_knot_surface(u, p, q, scale)
        pos2 = torus_knot_surface(u + 0.01, p, q, scale)
        # Two vectors along the torus' centerline
        vec1 = pos1 - pos2
        vec2 = pos1 + pos2
        # Two vectors orthoginal to the torus' centerline
        vec3 = np.cross(vec1, vec2)
        vec4 = np.cross(vec3, vec1)
        # Normalize
        vec3 /= ((vec3[:, 0] ** 2 + vec3[:, 1] ** 2 + vec3[:, 2] ** 2) ** 0.5).reshape(
            -1, 1
        )
        vec4 /= ((vec4[:, 0] ** 2 + vec4[:, 1] ** 2 + vec4[:, 2] ** 2) ** 0.5).reshape(
            -1, 1
        )
        # Define positions relative to the centerline
        cx = -tube * np.cos(v)
        cy = +tube * np.sin(v)
        # Prepare shapes, so we can do numpy broadcast
        pos = pos1.reshape(-1, 1, 3)
        cx.shape = 1, -1, 1
        cy.shape = 1, -1, 1
        vec3.shape = -1, 1, 3
        vec4.shape = -1, 1, 3
        # Broadcast!
        positions = pos + cx * vec4 + cy * vec3
        normals = positions - pos
        positions.shape = -1, 3
        normals.shape = -1, 3

        # Create texcords
        # ty, tx = np.meshgrid(u / u[-1], v / v[-1])
        ty, tx = np.meshgrid(v / v[-1], u / u[-1])
        texcoords = np.column_stack((tx.flat, ty.flat))

        # Create indices
        # Two triangles onto the "top-left" rectangle (six vertices)
        indices = np.array(
            [0, radial_verts, radial_verts + 1, radial_verts + 1, 1, 0],
            np.int32,
        )
        # Replicate to all rectangles, add offsets
        indices = np.tile(indices, (tubular_segments, radial_segments, 1))
        gx, gy = np.meshgrid(
            np.arange(indices.shape[1]), radial_verts * np.arange(indices.shape[0])
        )
        indices += (gx + gy).reshape(indices.shape[:2] + (1,))
        # Stitch the ends together over both axii.
        if stitch:
            indices[-1, :, 1:4] -= radial_verts * tubular_verts
            indices[:, -1, 2:5] -= radial_verts
        indices = indices.reshape(-1)

        # Create buffers for this geometry
        self.positions = Buffer(positions, usage="vertex|storage")
        self.normals = Buffer(normals, usage="vertex|storage")
        self.index = Buffer(indices, usage="index|storage")
        self.texcoords = Buffer(texcoords, usage="vertex|storage")


def torus_knot_surface(u, p, q, radius):
    """Taken from ThreeJS, but vectorized."""
    cu = np.cos(u)
    su = np.sin(u)
    qu_over_p = q / p * u
    cs = np.cos(qu_over_p)
    x = (2 + cs) * cu * (0.5 * radius)
    y = (2 + cs) * su * (0.5 * radius)
    z = np.sin(qu_over_p) * (0.5 * radius)
    return np.column_stack((x, y, z))



from ._base import WorldObject
from ..geometries import BoxGeometry
from ..resources import Buffer


class Volume(WorldObject):
    """A volume represents a 3D image in space. It has an implicit
    geometry based on the shape of the texture (the map of the
    material), and the center of voxel (0,0,0) will be at the origin
    of the local coordinate frame. Positioning and dealing with
    anisotropy should be dealt with using the scale and position
    properties.

    The picking info of a Volume (the result of
    ``renderer.get_pick_info()``) will for most materials include
    ``voxel_index`` (tuple of 3 floats).
    """

    def __init__(self, size, material):
        super().__init__()
        self.size = size
        self.material = material

    @property
    def material(self):
        """The material of the volume."""
        return self._material

    @material.setter
    def material(self, material):
        self._material = material

    @property
    def size(self):
        """The size of the volume (xyz)."""
        return self._size

    @size.setter
    def size(self, size):
        # Check and store
        x, y, z = size
        self._size = size = int(x), int(y), int(z)
        # Create box geometry, and map to 0..1
        geometry = BoxGeometry(1, 1, 1)
        geometry.positions.data[:, :3] += 0.5
        # This is our 3D texture coords
        geometry.texcoords = Buffer(
            geometry.positions.data[:, :3].copy(), usage="vertex|storage"
        )
        # Map to volume size
        for i in range(3):
            column = geometry.positions.data[:, i]
            column *= size[i]
            column -= 0.5  # Pixel centers are in origin and size

        # Apply
        self.geometry = geometry



# flake8: noqa

from . import utils

from .resources import *
from .objects import *
from .geometries import *
from .materials import *
from .cameras import *
from .helpers import *
from .controls import *

from .renderers import *


__version__ = "0.1.1"
version_info = tuple(map(int, __version__.split(".")))



class Matrix3:
    def __init__(self) -> None:
        self.elements = [
            1,
            0,
            0,
            0,
            1,
            0,
            0,
            0,
            1,
        ]

    def set(
        self,
        n11: float,
        n12: float,
        n13: float,
        n21: float,
        n22: float,
        n23: float,
        n31: float,
        n32: float,
        n33: float,
    ) -> "Matrix4":
        te = self.elements

        te[0] = n11
        te[3] = n12
        te[6] = n13
        te[1] = n21
        te[4] = n22
        te[7] = n23
        te[2] = n31
        te[5] = n32
        te[8] = n33

        return self



import numpy as np

from ..objects._base import ResourceContainer
from ..resources import Buffer


class Geometry(ResourceContainer):
    """A geometry represents the (input) data of mesh, line, or point
    geometry. It can include vertex positions, normals, colors, uvs,
    and custom data buffers. Face indices can be given using `index`.

    Buffer data can be provided as kwargs, these are converted to numpy arrays
    (if necessary) and wrapped in a Buffer.

    Example:

        g = Geometry(positions=[[1, 2], [2, 4], [3, 5], [4, 1]])
        g.positions.data  # numpy array

    """

    def __init__(self, **data):
        super().__init__()
        for name, val in data.items():
            if not isinstance(val, np.ndarray):
                val = np.asanyarray(val, dtype=np.float32)
            elif val.dtype == np.float64:
                raise ValueError(
                    "64-bit float is not supported, use 32-bit floats instead"
                )
            if name == "index":
                usage = "index|storage"
                # No assumptions about shape; they're considered flat anyway
            elif name == "positions":
                usage = "vertex|storage"
                if not (val.ndim == 2 and val.shape[-1] == 3):
                    raise ValueError("Expected Nx3 data for positions")
            elif name == "normals":
                usage = "vertex|storage"
                if not (val.ndim == 2 and val.shape[-1] == 3):
                    raise ValueError("Expected Nx3 data for normals")
            elif name == "texcoords":
                usage = "vertex|storage"
                if val.ndim == 1:
                    pass
                elif not (val.ndim == 2 and val.shape[-1] in (1, 2, 3)):
                    raise ValueError("Expected Nx1, Nx2 or Nx3 data for texcoords")
            else:
                usage = "vertex|storage"
            setattr(self, name, Buffer(val, usage=usage))



from ._base import WorldObject


class Line(WorldObject):
    """An object consisting of a line represented by a list of vertices
    (3D positions). Some materials will render the line as a continuous line,
    while other materials will consider each pair of points a segment.

    The picking info of a Line (the result of
    ``renderer.get_pick_info()``) will for most materials include
    ``vertex_index`` (float). Note that ``vertex_index`` is not integer;
    round the number to obtain the nearest vertex.
    """

    def __init__(self, geometry, material):
        super().__init__()
        self.geometry = geometry
        self.material = material



import numpy as np

from .. import Geometry, Line, LineThinSegmentMaterial


DTYPE = "f4"


class GridHelper(Line):
    def __init__(
        self,
        size=10.0,
        divisions=10,
        color1=(0.35, 0.35, 0.35, 1),
        color2=(0.1, 0.1, 0.1, 1),
    ):
        assert isinstance(divisions, int)
        assert size > 0.0

        half_size = size / 2
        n_lines = divisions + 1
        x = np.linspace(-half_size, half_size, num=n_lines, dtype=DTYPE)

        # the grid is made up of 2 * n_lines line segments
        # where each line has two endpoints (2, 3)
        positions = np.zeros((2, n_lines, 2, 3), dtype=DTYPE)
        positions[0, ..., 0] = x[:, None]
        positions[0, ..., 2] = [[-half_size, half_size]]
        positions[1, ..., 0] = [[-half_size, half_size]]
        positions[1, ..., 2] = x[:, None]

        # color1 for the center lines, color2 for the rest
        colors = np.empty((2, n_lines, 2, 4), dtype=DTYPE)
        colors[..., :] = color2
        colors[:, n_lines // 2, :, :] = color1

        geometry = Geometry(
            positions=positions.reshape((-1, 3)), colors=colors.reshape((-1, 4))
        )
        material = LineThinSegmentMaterial(vertex_colors=True)

        super().__init__(geometry, material)



from ._base import Camera
from ..linalg import Matrix4


class OrthographicCamera(Camera):
    """An orthographic camera, useful for non-perspective views and
    visualizing 2D content. You may also want to set the position of
    the camera.

    Parameters:
        width (float): The (minimum) width of the view-cube. The actual view
           may be wider if the viewport is wide.
        height (float): The (minimum) height of the view-cube.
        near (float): The near clipping plane. Default -1000.
        far (float): The far clipping plane. Must be larger than near. Default +1000.
    """

    def __init__(self, width=1, height=1, near=-1000, far=1000):
        super().__init__()
        # These width and height represent the view-plane in world coordinates
        # and has little to do with the canvas/viewport size.
        self.width = float(width)
        self.height = float(height)
        self.near = float(near)
        self.far = float(far)
        assert self.near < self.far
        self.zoom = 1
        self.maintain_aspect = True
        self.set_viewport_size(1, 1)
        self.update_projection_matrix()

    def __repr__(self) -> str:
        return (
            f"OrthographicCamera({self.width}, {self.height}, {self.near}, {self.far})"
        )

    def set_viewport_size(self, width, height):
        self._view_aspect = width / height
        # The reference view plane is scaled with the zoom factor
        width = self.width / self.zoom
        height = self.height / self.zoom
        # Increase either the width or height, depending on the viewport shape
        aspect = width / height
        if not self.maintain_aspect:
            pass
        elif aspect < self._view_aspect:
            width *= self._view_aspect / aspect
        else:
            height *= aspect / self._view_aspect
        # Store the size if the visible view plane in world coordinates
        self.visible_world_size = width, height

    def update_projection_matrix(self):
        w, h = self.visible_world_size
        bottom = -0.5 * h
        top = +0.5 * h
        left = -0.5 * w
        right = +0.5 * w
        # Set matrices
        # The linalg ortho projection puts xyz in the range -1..1, but
        # in the coordinate system of wgpu (and this lib) the depth is
        # expressed in 0..1, so we also correct for that.
        self.projection_matrix.make_orthographic(
            left, right, top, bottom, self.near, self.far
        )
        self.projection_matrix.premultiply(
            Matrix4(1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0.5, 0, 0, 0, 0.5, 1)
        )
        self.projection_matrix_inverse.get_inverse(self.projection_matrix)



import wgpu  # only for flags/enums

from . import register_wgpu_render_function
from ._shadercomposer import BaseShader
from ...objects import Points
from ...materials import PointsMaterial, GaussianPointsMaterial


@register_wgpu_render_function(Points, PointsMaterial)
def points_renderer(wobject, render_info):
    """Render function capable of rendering meshes displaying a volume slice."""

    geometry = wobject.geometry
    material = wobject.material
    shader = PointsShader(type="circle")
    n = geometry.positions.nitems * 6

    shader.define_uniform(0, 0, "u_stdinfo", render_info.stdinfo_uniform.data.dtype)
    shader.define_uniform(0, 1, "u_wobject", wobject.uniform_buffer.data.dtype)
    shader.define_uniform(0, 2, "u_material", material.uniform_buffer.data.dtype)

    # Collect bindings
    bindings0 = {
        0: ("buffer/uniform", render_info.stdinfo_uniform),
        1: ("buffer/uniform", wobject.uniform_buffer),
        2: ("buffer/uniform", material.uniform_buffer),
    }

    bindings1 = {
        0: ("buffer/read_only_storage", geometry.positions),
    }

    if isinstance(material, GaussianPointsMaterial):
        shader["type"] = "gaussian"

    # Put it together!
    wgsl = shader.generate_wgsl()
    return [
        {
            "vertex_shader": (wgsl, "vs_main"),
            "fragment_shader": (wgsl, "fs_main"),
            "primitive_topology": wgpu.PrimitiveTopology.triangle_list,
            "indices": (range(n), range(1)),
            "vertex_buffers": {},
            "bindings0": bindings0,
            "bindings1": bindings1,
        }
    ]


class PointsShader(BaseShader):

    # Notes:
    # In WGPU, the pointsize attribute can no longer be larger than 1 because
    # of restriction in some hardware/backend API's. So we use our storage-buffer
    # approach (similar for what we use for lines) to sort of fake a geometry shader.
    # An alternative is to use instancing. Could be worth testing both approaches
    # for performance ...

    def get_code(self):
        return (
            self.get_definitions()
            + self.more_definitions()
            + self.vertex_shader()
            + self.fragment_shader()
        )

    def more_definitions(self):
        return """

        struct VertexInput {
            [[builtin(vertex_index)]] index : u32;
        };
        struct VertexOutput {
            [[location(0)]] pointcoord: vec2<f32>;
            [[location(1)]] vertex_idx: vec2<f32>;
            [[builtin(position)]] pos: vec4<f32>;
        };

        struct FragmentOutput {
            [[location(0)]] color: vec4<f32>;
            [[location(1)]] pick: vec4<i32>;
        };

        [[block]]
        struct BufferF32 {
            data: [[stride(4)]] array<f32>;
        };

        [[group(1), binding(0)]]
        var<storage> s_pos: [[access(read)]] BufferF32;
        """

    def vertex_shader(self):
        return """

        [[stage(vertex)]]
        fn vs_main(in: VertexInput) -> VertexOutput {
            var out: VertexOutput;

            let index = i32(in.index);
            let i0 = index / 6;
            let sub_index = index % 6;

            let raw_pos = vec3<f32>(s_pos.data[i0 * 3 + 0], s_pos.data[i0 * 3 + 1], s_pos.data[i0 * 3 + 2]);
            let world_pos = u_wobject.world_transform * vec4<f32>(raw_pos, 1.0);
            let ndc_pos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;

            let deltas = array<vec2<f32>, 6>(
                vec2<f32>(-1.0, -1.0),
                vec2<f32>(-1.0,  1.0),
                vec2<f32>( 1.0, -1.0),
                vec2<f32>(-1.0,  1.0),
                vec2<f32>( 1.0, -1.0),
                vec2<f32>( 1.0,  1.0),
            );

            let aa_margin = 1.0;
            let delta_logical = deltas[sub_index] * (u_material.size + aa_margin);
            let delta_ndc = delta_logical * (1.0 / u_stdinfo.logical_size);
            out.pos = vec4<f32>(ndc_pos.xy + delta_ndc, ndc_pos.zw);
            out.pointcoord = delta_logical;

            out.vertex_idx = vec2<f32>(f32(i0 / 10000), f32(i0 % 10000));
            return out;
        }
        """

    def fragment_shader(self):
        # Also see See https://github.com/vispy/vispy/blob/master/vispy/visuals/markers.py
        return """

        [[stage(fragment)]]
        fn fs_main(in: VertexOutput) -> FragmentOutput {
            var out: FragmentOutput;

            let color = u_material.color;
            let d = length(in.pointcoord);
            let size = u_material.size;
            let aa_width = 1.0;

            $$ if type == 'circle'
                if (d <= size - 0.5 * aa_width) {
                    out.color = color;
                } elseif (d <= size + 0.5 * aa_width) {
                    let alpha1 = 0.5 + (size - d) / aa_width;
                    let alpha2 = pow(alpha1, 2.0);  // this works better
                    out.color = vec4<f32>(color.rgb, color.a * alpha2);
                } else {
                    discard;
                }
            $$ elif type == "gaussian"
                if (d <= size) {
                    let sigma = size / 3.0;
                    let t = d / sigma;
                    let a = exp(-0.5 * t * t);
                    out.color = vec4<f32>(color.rgb, color.a * a);
                } else {
                    discard;
                }
            $$ else
                invalid_point_type;
            $$ endif

            // Picking
            let vertex_id = i32(in.vertex_idx.x * 10000.0 + in.vertex_idx.y + 0.5);
            out.pick = vec4<i32>(u_wobject.id, 0, vertex_id, 0);

            return out;
        }
        """



import wgpu  # only for flags/enums

from . import register_wgpu_render_function
from ._shadercomposer import BaseShader
from ...objects import Background
from ...materials import BackgroundMaterial, BackgroundImageMaterial
from ...resources import Texture, TextureView


@register_wgpu_render_function(Background, BackgroundMaterial)
def background_renderer(wobject, render_info):

    material = wobject.material
    shader = BackgroundShader(texture_dim="")

    bindings0 = {
        0: ("buffer/uniform", render_info.stdinfo_uniform),
        1: ("buffer/uniform", wobject.uniform_buffer),
        2: ("buffer/uniform", material.uniform_buffer),
    }

    shader.define_uniform(0, 0, "u_stdinfo", render_info.stdinfo_uniform.data.dtype)
    shader.define_uniform(0, 1, "u_wobject", wobject.uniform_buffer.data.dtype)
    shader.define_uniform(0, 2, "u_material", material.uniform_buffer.data.dtype)

    bindings1 = {}

    if isinstance(material, BackgroundImageMaterial) and material.map is not None:
        if isinstance(material.map, Texture):
            raise TypeError("material.map is a Texture, but must be a TextureView")
        elif not isinstance(material.map, TextureView):
            raise TypeError("material.map must be a TextureView")
        bindings1[0] = "sampler/filtering", material.map
        bindings1[1] = "texture/auto", material.map
        # Select shader
        if material.map.view_dim == "cube":
            shader["texture_dim"] = "cube"
        elif material.map.view_dim == "2d":
            shader["texture_dim"] = "2d"
        else:
            raise ValueError(
                "BackgroundImageMaterial should have map with texture view 2d or cube."
            )
        # Channels
        if material.map.format.startswith("rgb"):  # rgb maps to rgba
            shader["texture_color"] = True
        elif material.map.format.startswith("r"):
            shader["texture_color"] = False
        else:
            raise ValueError("Unexpected texture format")

    wgsl = shader.generate_wgsl()
    return [
        {
            "vertex_shader": (wgsl, "vs_main"),
            "fragment_shader": (wgsl, "fs_main"),
            "primitive_topology": wgpu.PrimitiveTopology.triangle_strip,
            "indices": 4,
            "bindings0": bindings0,
            "bindings1": bindings1,
        }
    ]


class BackgroundShader(BaseShader):
    def get_code(self):
        return (
            self.get_definitions()
            + self.more_definitions()
            + self.vertex_shader()
            + self.fragment_shader()
        )

    def more_definitions(self):
        return """

        struct VertexInput {
            [[builtin(vertex_index)]] index : u32;
        };
        struct VertexOutput {
            [[location(0)]] texcoord: vec3<f32>;
            [[builtin(position)]] pos: vec4<f32>;
        };

        struct FragmentOutput {
            [[location(0)]] color: vec4<f32>;
            [[location(1)]] pick: vec4<i32>;
        };

        $$ if texture_dim
        [[group(1), binding(0)]]
        var r_sampler: sampler;

        [[group(1), binding(1)]]
        var r_tex: texture_{{ texture_dim }}<f32>;
        $$ endif
    """

    def vertex_shader(self):
        return """
        [[stage(vertex)]]
        fn vs_main(in: VertexInput) -> VertexOutput {
            var out: VertexOutput;
            // Define positions at the four corners of the viewport, at the largest depth
            let positions = array<vec2<f32>, 4>(
                vec2<f32>(-1.0, -1.0),
                vec2<f32>( 1.0, -1.0),
                vec2<f32>(-1.0,  1.0),
                vec2<f32>( 1.0,  1.0),
            );
            // Select the current position
            let pos = positions[i32(in.index)];
            $$ if texture_dim == "cube"
                let ndc_pos1 = vec4<f32>(pos, 0.9999999, 1.0);
                let ndc_pos2 = vec4<f32>(pos, 1.1000000, 1.0);
                // project back to world coords
                let ndc_to_world = u_stdinfo.cam_transform_inv * u_stdinfo.projection_transform_inv;
                //let ndc_to_world = u_stdinfo.ndc_to_world;
                let wpos1_ = ndc_to_world * ndc_pos1;
                let wpos2_ = ndc_to_world * ndc_pos2;
                let wpos1 = wpos1_.xyzw / wpos1_.w;
                let wpos2 = wpos2_.xyzw / wpos2_.w;
                // Store positions and the view direction in the world
                out.pos = ndc_pos1;
                out.texcoord = wpos2.xyz - wpos1.xyz;
            $$ else
                // Store positions and the view direction in the world
                out.pos = vec4<f32>(pos, 0.9999999, 1.0);;
                out.texcoord = vec3<f32>(pos * 0.5 + 0.5, 0.0);
            $$ endif
            return out;
        }
        """

    def fragment_shader(self):
        return """
        [[stage(fragment)]]
        fn fs_main(in: VertexOutput) -> FragmentOutput {
            var out: FragmentOutput;
            $$ if texture_dim
                $$ if texture_dim == '2d'
                    let color = textureSample(r_tex, r_sampler, in.texcoord.xy);
                $$ elif texture_dim == 'cube'
                    let color = textureSample(r_tex, r_sampler, in.texcoord.xyz);
                $$ endif
                $$ if texture_color
                    out.color = color.rgba;
                $$ else
                    out.color = color.rrra;
                $$ endif
            $$ else
                let f = in.texcoord.xy;
                out.color = (
                    u_material.color_bottom_left * (1.0 - f.x) * (1.0 - f.y)
                    + u_material.color_bottom_right * f.x * (1.0 - f.y)
                    + u_material.color_top_left * (1.0 - f.x) * f.y
                    + u_material.color_top_right * f.x * f.y
                );
            $$ endif
            return out;
        }
        """



import wgpu  # only for flags/enums

from . import register_wgpu_render_function
from ._shadercomposer import BaseShader
from ...utils import array_from_shadertype
from ...resources import Buffer
from ...objects import Line
from ...materials import (
    LineMaterial,
    LineThinMaterial,
    LineThinSegmentMaterial,
    LineSegmentMaterial,
    LineArrowMaterial,
)


# ## Notes
#
# Rendering lines on the GPU is hard! The approach used here uses
# VertexId and storage buffers instead of vertex buffers. That way we
# can create multiple vertices for each point on the line. These
# vertices are positioned such that rendering with triangle_strip
# results in a tick line. More info below.
#
# An alternative (used by e.g. ThreeJS) is to draw the geometry of a
# line segment many times using instancing. Note that some Intel drivers
# limit the number of instances to about 5-10k, which is not much considering
# that lines consisting of millions of points can normally run realtime.
# https://wwwtyro.net/2019/11/18/instanced-lines.html
#
# Another alternative is a geometry shader, but wgpu does not have that.
# But we have compute shaders, which can be used to prepare the triangle-based
# geometry and store it in a buffer, which can be used as a vertex buffer in
# the render pass. The downside of this approach is the memory consumption.
# But a hybrid approach may be viable: the current approach using VertexId,
# but preparing some data in a buffer.

# todo: we can learn about dashing, unfolding and more at http://jcgt.org/published/0002/02/08/paper.pdf


renderer_uniform_type = dict(last_i=("int32",))


@register_wgpu_render_function(Line, LineMaterial)
def line_renderer(wobject, render_info):
    """Render function capable of rendering lines."""

    material = wobject.material
    geometry = wobject.geometry
    shader = LineShader()

    assert isinstance(material, LineMaterial)

    positions1 = geometry.positions

    # With vertex buffers, if a shader input is vec4, and the vbo has
    # Nx2, the z and w element will be zero. This works, because for
    # vertex buffers we provide additional information about the
    # striding of the data.
    # With storage buffers (aka SSBO) we just have some bytes that we
    # read from/write to in the shader. This is more free, but it means
    # that the data in the buffer must match with what the shader
    # expects. In addition to that, there's this thing with vec3's which
    # are padded to 16 bytes. So we either have to require our users
    # to provide Nx4 data, or read them as an array of f32.
    # Anyway, extra check here to make sure the data matches!
    # todo: data.something in here, which means we assume numpy-ish arrays
    if positions1.data.shape[1] != 3:
        raise ValueError(
            "For rendering (thick) lines, the geometry.positions must be Nx3."
        )

    uniform_buffer = Buffer(
        array_from_shadertype(renderer_uniform_type), usage="UNIFORM"
    )
    uniform_buffer.data["last_i"] = positions1.nitems - 1

    shader.define_uniform(0, 0, "u_stdinfo", render_info.stdinfo_uniform.data.dtype)
    shader.define_uniform(0, 1, "u_wobject", wobject.uniform_buffer.data.dtype)
    shader.define_uniform(0, 2, "u_material", material.uniform_buffer.data.dtype)
    shader.define_uniform(0, 3, "u_renderer", uniform_buffer.data.dtype)

    bindings0 = {
        0: ("buffer/uniform", render_info.stdinfo_uniform),
        1: ("buffer/uniform", wobject.uniform_buffer),
        2: ("buffer/uniform", material.uniform_buffer),
        3: ("buffer/uniform", uniform_buffer),
    }

    bindings1 = {
        0: ("buffer/read_only_storage", positions1),
    }

    # Global color or per-vertex?
    shader["per_vertex_color"] = False
    if material.vertex_colors:
        colors1 = geometry.colors
        if colors1.data.shape[1] != 4:
            raise ValueError(
                "For rendering (thick) lines, the geometry.colors must be Nx4."
            )
        bindings1[1] = ("buffer/read_only_storage", colors1)
        shader["per_vertex_color"] = True

    if isinstance(material, LineArrowMaterial):
        shader["line_type"] = "arrow"
        n = (positions1.nitems // 2) * 2 * 4
    elif isinstance(material, LineSegmentMaterial):
        shader["line_type"] = "segment"
        n = (positions1.nitems // 2) * 2 * 5
    else:
        shader["line_type"] = "line"
        n = positions1.nitems * 5

    # Done
    wgsl = shader.generate_wgsl()
    return [
        {
            "vertex_shader": (wgsl, "vs_main"),
            "fragment_shader": (wgsl, "fs_main"),
            "primitive_topology": wgpu.PrimitiveTopology.triangle_strip,
            "indices": (n, 1),
            "bindings0": bindings0,
            "bindings1": bindings1,
            "target": None,  # default
        },
    ]


class LineShader(BaseShader):
    def get_code(self):
        return (
            self.get_definitions()
            + self.more_definitions()
            + self.helpers()
            + self.vertex_shader()
            + self.fragment_shader()
        )

    def more_definitions(self):
        return """

        let PI: f32 = 3.14159265359;

        struct VertexInput {
            [[builtin(vertex_index)]] index : u32;
        };
        struct VertexOutput {
            [[location(0)]] line_width_p: f32;
            [[location(1)]] vec_from_node_p: vec2<f32>;
            [[location(2)]] color: vec4<f32>;
            [[location(3)]] vertex_idx: vec2<f32>;
            [[builtin(position)]] pos: vec4<f32>;
        };

        struct VertexFuncOutput {
            i: i32;
            pos: vec4<f32>;
            line_width_p: f32;
            vec_from_node_p: vec2<f32>;
        };

        struct FragmentOutput {
            [[location(0)]] color: vec4<f32>;
            [[location(1)]] pick: vec4<i32>;
            [[builtin(frag_depth)]] depth : f32;
        };

        [[block]]
        struct BufferI32 {
            data: [[stride(4)]] array<i32>;
        };

        [[block]]
        struct BufferF32 {
            data: [[stride(4)]] array<f32>;
        };

        // Could be useful to be able to pass via an index too :)
        // [[group(1), binding(0)]]
        // var<storage> s_indices: [[access(read)]] BufferI32;

        [[group(1), binding(0)]]
        var<storage> s_pos: [[access(read)]] BufferF32;

        $$ if per_vertex_color
        [[group(1), binding(1)]]
        var<storage> s_color: [[access(read)]] BufferF32;
        $$ endif
        """

    def helpers(self):
        return """

        fn get_point_ndc(index:i32) -> vec4<f32> {
            let raw_pos = vec3<f32>(
                s_pos.data[index], s_pos.data[index + 1], s_pos.data[index + 2]
            );
            let world_pos = u_wobject.world_transform * vec4<f32>(raw_pos.xyz, 1.0);
            var ndc_pos: vec4<f32> = u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;
            ndc_pos = ndc_pos / ndc_pos.w;
            return ndc_pos;
        }
        """

    def vertex_shader(self):
        if self.kwargs["line_type"] == "arrow":
            core = self.vertex_shader_arrow()
        elif self.kwargs["line_type"] == "segment":
            core = self.vertex_shader_segment()
        else:
            core = self.vertex_shader_line()

        return (
            core
            + """
        [[stage(vertex)]]
        fn vs_main(in: VertexInput) -> VertexOutput {

            let index = i32(in.index);
            let screen_factor = u_stdinfo.logical_size.xy / 2.0;
            let half_line_width:f32 = u_material.thickness * 0.5;  // in logical pixels
            let l2p:f32 = u_stdinfo.physical_size.x / u_stdinfo.logical_size.x;

            let result: VertexFuncOutput = get_vertex_result(index, screen_factor, half_line_width, l2p);

            var out: VertexOutput;
            out.pos = result.pos;
            out.line_width_p = result.line_width_p;
            out.vec_from_node_p = result.vec_from_node_p;
            out.vertex_idx = vec2<f32>(f32(result.i / 10000), f32(result.i % 10000));

            $$ if per_vertex_color
            let i = result.i;
            out.color = vec4<f32>(
                s_color.data[i * 4], s_color.data[i * 4 + 1], s_color.data[i * 4 + 2], s_color.data[i * 4 + 3]
            );
            $$ else:
                out.color = u_material.color;
            $$ endif

            return out;
        }
        """
        )

    def vertex_shader_line(self):
        return """
        fn get_vertex_result(
            index:i32, screen_factor:vec2<f32>, half_line_width:f32, l2p:f32
        ) -> VertexFuncOutput {
            // This vertex shader uses VertexId and storage buffers instead of
            // vertex buffers. It creates 5 vertices for each point on the line.
            // You could do with 4 to get bevel joins, but round joins are so
            // much nicer. The extra vertex is used to cover more fragments at
            // the joins (and caps). In the fragment shader we discard fragments
            // that are "out of range", based on a varying that represents the
            // vector from the node to the vertex.
            //
            // Basic algorithm and definitions:
            //
            // - We read the positions of three nodes, the current, previous, and next.
            // - These are converted to logical pixel screen space.
            // - We define four normal vectors (na, nb, nc, nd) which represent the
            //   vertices. Two for the previous segment, two for the next. One extra
            //   normal/vertex is defined at the join.
            // - These calculations are done for each vertex (yeah, bit of a waste),
            //   we select just one as output.
            //
            //            /  o     node 3
            //           /  /  /
            //          d  /  /
            //   - - - b  /  /     corners rectangles a, b, c, d
            //   o-------o  /      the vertex e is extra
            //   - - - - a c
            //                node 2
            //  node 1
            //
            // Note that at the inside of a join, the normals (b and d above)
            // move past each-other, causing the rectangles of both segments to
            // overlap. We could prevent this, but that would screw up
            // v_vec_from_node_p. Furthermore, consider a thick line with dense
            // nodes jumping all over the place, causing a lot of overlap anyway.
            // In summary, this viz relies on depth testing with "less" (or
            // semi-tranparent lines would break).
            //
            // Possible improvents:
            //
            // - can we do dashes/stipling?
            // - also implement bevel and miter joins
            // - also implement different caps
            // - we can prepare the nodes' screen coordinates in a compute shader.

            let i = index / 5;

            // Sample the current node and it's two neighbours, and convert to NDC
            let npos1 = get_point_ndc(i * 3 - 3);
            let npos2 = get_point_ndc(i * 3 + 0);
            let npos3 = get_point_ndc(i * 3 + 3);

            // Convert to logical screen coordinates, because that's were the lines work
            let ppos1 = (npos1.xy + 1.0) * screen_factor;
            let ppos2 = (npos2.xy + 1.0) * screen_factor;
            let ppos3 = (npos3.xy + 1.0) * screen_factor;

            // Get vectors normal to the line segments
            var v1: vec2<f32> = ppos2.xy - ppos1.xy;
            var v2: vec2<f32> = ppos3.xy - ppos2.xy;
            var na: vec2<f32>;
            var nb: vec2<f32>;
            var nc: vec2<f32>;
            var nd: vec2<f32>;
            var ne: vec2<f32>;

            if (i == 0) {
                // This is the first point on the line: create a cap.
                v1 = v2;
                nc = normalize(vec2<f32>(v2.y, -v2.x));
                nd = -nc;
                na = nd;
                nb = nd - normalize(v2);
                ne = nc - normalize(v2);
            } elseif (i == u_renderer.last_i) {
                // This is the last point on the line: create a cap.
                v2 = v1;
                na = normalize(vec2<f32>(v1.y, -v1.x));
                nb = -na;
                ne = nb + normalize(v1);
                nc = na + normalize(v1);
                nd = na;
            } else {
                // Create a join
                na = normalize(vec2<f32>(v1.y, -v1.x));
                nb = -na;
                nc = normalize(vec2<f32>(v2.y, -v2.x));
                nd = -nc;

                // Determine the angle between two of the normals. If this angle is smaller
                // than zero, the inside of the join is at nb/nd, otherwise it is at na/nc.
                var angle:f32 = atan2(na.y, na.x) - atan2(nc.y, nc.x);
                angle = (angle + PI) % (2.0 * PI) - PI;

                // From the angle we can also calculate the intersection of the lines.
                // We express it in a vector magnifier, and limit it to a factor 2,
                // since when the angle is ~pi, the intersection is near infinity.
                // For a bevel join we can omit ne (or set vec_mag to 1.0).
                // For a miter join we'd need an extra vertex to smoothly transition
                // from a miter to a bevel when the angle is too small.
                // Note that ne becomes inf if v1 == v2, but that's ok, because the
                // triangles in which ne takes part are degenerate for this use-case.
                let vec_mag = 1.0 / max(0.25, cos(0.5 * angle));
                ne = normalize(normalize(v1) - normalize(v2)) * vec_mag;
            }

            // Select the correct vector, note that all except ne are unit.
            let vectors = array<vec2<f32>,5>(na, nb, ne, nc, nd);
            let the_vec = vectors[index % 5] * half_line_width;

            var out : VertexFuncOutput;
            out.i = i;
            out.pos = vec4<f32>((ppos2 + the_vec) / screen_factor - 1.0, npos2.zw);
            out.line_width_p = half_line_width * 2.0 * l2p;
            out.vec_from_node_p = the_vec * l2p;
            return out;
        }
        """

    def vertex_shader_segment(self):
        return """
        fn get_vertex_result(
            index:i32, screen_factor:vec2<f32>, half_line_width:f32, l2p:f32
        ) -> VertexFuncOutput {
            // Similar to the regular line shader, except we only draw segments,
            // using 5 vertices per node. Four for the segments, and 1 to create
            // a degenerate triangle for the space in between. So we only draw
            // caps, no joins.

            let i = index / 5;

            // Sample the current node and either of its neighbours
            let i3 = i + 1 - (i % 2) * 2;  // (i + 1) if i is even else (i - 1)
            let npos2 = get_point_ndc(i * 3);
            let npos3 = get_point_ndc(i3 * 3 );
            // Convert to logical screen coordinates, because that's were the lines work
            let ppos2 = (npos2.xy + 1.0) * screen_factor;
            let ppos3 = (npos3.xy + 1.0) * screen_factor;

            // Get vectors normal to the line segments
            var na: vec2<f32>;
            var nb: vec2<f32>;
            var nc: vec2<f32>;
            var nd: vec2<f32>;

            // Get vectors normal to the line segments
            if ((i % 2) == 0) {
                // A left-cap
                let v = normalize(ppos3.xy - ppos2.xy);
                nc = vec2<f32>(v.y, -v.x);
                nd = -nc;
                na = nc - v;
                nb = nd - v;
            } else {
                // A right cap
                let v = normalize(ppos2.xy - ppos3.xy);
                na = vec2<f32>(v.y, -v.x);
                nb = -na;
                nc = na + v;
                nd = nb + v;
            }

            // Select the correct vector
            // Note the replicated vertices to create degenerate triangles
            let vectors = array<vec2<f32>,10>(na, na, nb, nc, nd, na, nb, nc, nd, nd);
            let the_vec = vectors[index % 10] * half_line_width;

            var out : VertexFuncOutput;
            out.i = i;
            out.pos = vec4<f32>((ppos2 + the_vec) / screen_factor - 1.0, npos2.zw);
            out.line_width_p = half_line_width * 2.0 * l2p;
            out.vec_from_node_p = the_vec * l2p;
            return out;
        }
        """

    def vertex_shader_arrow(self):
        return """
        fn get_vertex_result(
            index:i32, screen_factor:vec2<f32>, half_line_width:f32, l2p:f32
        ) -> VertexFuncOutput {
            // Similar to the normal vertex shader, except we only draw segments,
            // using 3 vertices per node: 6 per segment. 4 for the arrow, and 2
            // to create a degenerate triangle for the space in between. So we
            // only draw caps, no joins.

            let i = index / 3;

            // Sample the current node and either of its neighbours
            let i3 = i + 1 - (i % 2) * 2;  // (i + 1) if i is even else (i - 1)
            let npos2 = get_point_ndc(i * 3);
            let npos3 = get_point_ndc(i3 * 3 );
            // Convert to logical screen coordinates, because that's were the lines work
            let ppos2 = (npos2.xy + 1.0) * screen_factor;
            let ppos3 = (npos3.xy + 1.0) * screen_factor;

            // Get vectors normal to the line segments
            var na: vec2<f32>;
            var nb: vec2<f32>;

            // Get vectors normal to the line segments
            if ((i % 2) == 0) {
                // A left-cap
                let v = ppos3.xy - ppos2.xy;
                na = normalize(vec2<f32>(v.y, -v.x)) * half_line_width;
                nb = v;
            } else {
                // A right cap
                let v = ppos2.xy - ppos3.xy;
                na = -0.75 * v;
                nb = normalize(vec2<f32>(-v.y, v.x)) * half_line_width - v;
            }

            // Select the correct vector
            // Note the replicated vertices to create degenerate triangles
            let vectors = array<vec2<f32>,6>(na, na, nb, na, nb, nb);
            let the_vec = vectors[index % 6];

            var out : VertexFuncOutput;
            out.i = i;
            out.pos = vec4<f32>((ppos2 + the_vec) / screen_factor - 1.0, npos2.zw);
            out.line_width_p = half_line_width * 2.0 * l2p;
            out.vec_from_node_p = vec2<f32>(0.0, 0.0);
            return out;
        }
        """

    def fragment_shader(self):
        return """
        [[stage(fragment)]]
        fn fs_main(in: VertexOutput) -> FragmentOutput {

            // Discard fragments outside of the radius. This is what makes round
            // joins and caps. If we ever want bevel or miter joins, we should
            // change the vertex positions a bit, and drop these lines below.
            let dist_to_node_p = length(in.vec_from_node_p);
            if (dist_to_node_p > in.line_width_p * 0.5) {
                discard;
            }

            // Prep
            var out: FragmentOutput;
            var alpha: f32 = 1.0;

            // Anti-aliasing. Note that because of the discarding above, we cannot
            // use MSAA for aa. But maybe we use another generic approach to aa. We'll see.
            // todo: because of this, our line gets a wee bit thinner, so we have to
            // output ticker lines in the vertex shader!
            let aa_width = 1.2;
            alpha = ((0.5 * in.line_width_p) - abs(dist_to_node_p)) / aa_width;
            alpha = pow(min(1.0, alpha), 2.0);

            // The outer edges with lower alpha for aa are pushed a bit back to avoid artifacts.
            out.depth = in.pos.z + 0.0001 * (0.8 - min(0.8, alpha));

            // Set color
            let color = in.color;
            out.color = vec4<f32>(color.rgb, min(1.0, color.a) * alpha);

            // Set picking info. Yes, the vertex_id interpolates correctly in encoded form.
            let vf: f32 = in.vertex_idx.x * 10000.0 + in.vertex_idx.y;
            let vi = i32(vf + 0.5);
            out.pick = vec4<i32>(u_wobject.id, 0, vi, i32((vf - f32(vi)) * 1048576.0));

            return out;
        }
        """


@register_wgpu_render_function(Line, LineThinSegmentMaterial)
@register_wgpu_render_function(Line, LineThinMaterial)
def thin_line_renderer(wobject, render_info):
    """Render function capable of rendering lines."""

    material = wobject.material
    geometry = wobject.geometry
    shader = ThinLineShader()

    positions1 = geometry.positions

    primitive = wgpu.PrimitiveTopology.line_strip
    if isinstance(material, LineThinSegmentMaterial):
        primitive = wgpu.PrimitiveTopology.line_list

    shader.define_uniform(0, 0, "u_stdinfo", render_info.stdinfo_uniform.data.dtype)
    shader.define_uniform(0, 1, "u_wobject", wobject.uniform_buffer.data.dtype)
    shader.define_uniform(0, 2, "u_material", material.uniform_buffer.data.dtype)

    bindings0 = {
        0: ("buffer/uniform", render_info.stdinfo_uniform),
        1: ("buffer/uniform", wobject.uniform_buffer),
        2: ("buffer/uniform", material.uniform_buffer),
    }

    vertex_buffers = {0: positions1}

    shader["per_vertex_color"] = False
    if material.vertex_colors:
        shader["per_vertex_color"] = True
        colors1 = geometry.colors
        if colors1.data.shape[1] != 4:
            raise ValueError(
                "For rendering (thick) lines, the geometry.colors must be Nx4."
            )
        vertex_buffers[1] = colors1

    wgsl = shader.generate_wgsl()
    return [
        {
            "vertex_shader": (wgsl, "vs_main"),
            "fragment_shader": (wgsl, "fs_main"),
            "primitive_topology": primitive,
            "indices": (positions1.nitems, 1),
            "vertex_buffers": vertex_buffers,
            "bindings0": bindings0,
        },
    ]


class ThinLineShader(BaseShader):
    def get_code(self):
        return (
            self.get_definitions()
            + self.more_definitions()
            + self.vertex_shader()
            + self.fragment_shader()
        )

    def more_definitions(self):
        return """

        struct VertexInput {
            [[location(0)]] pos : vec3<f32>;
            $$ if per_vertex_color
            [[location(1)]] color : vec4<f32>;
            $$ endif
            [[builtin(vertex_index)]] index : u32;
        };
        struct VertexOutput {
            $$ if per_vertex_color
            [[location(0)]] color: vec4<f32>;
            $$ endif
            [[builtin(position)]] pos: vec4<f32>;
        };

        struct FragmentOutput {
            [[location(0)]] color: vec4<f32>;
            [[location(1)]] pick: vec4<i32>;
        };
        """

    def vertex_shader(self):
        return """
        [[stage(vertex)]]
        fn vs_main(in: VertexInput) -> VertexOutput {

            let wpos = u_wobject.world_transform * vec4<f32>(in.pos.xyz, 1.0);
            let npos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * wpos;

            var out: VertexOutput;
            out.pos = npos;
            $$ if per_vertex_color
            out.color = in.color;
            $$ endif
            return out;
        }
        """

    def fragment_shader(self):
        return """
        [[stage(fragment)]]
        fn fs_main(in: VertexOutput) -> FragmentOutput {
            var out : FragmentOutput;
            $$ if per_vertex_color
            out.color = in.color;
            $$ else
            out.color = u_material.color;
            $$ endif
            return out;
        }
        """
